#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import numexpr
import os
import uuid

from string import Template
from numpy import array2string
from typing import Callable
from datetime import datetime
from getgauge.python import data_store


def substitute(gauge_param: str) -> str:
    """Substitutes placeholders in a step parameter with values from environment variables
    and evaluates mathematical expressions.
    The environment variables are usually defined in the env/*.properties files in the gauge project
    or are placed into the context with specific steps.
    So the same placeholder can be replaced with different values in different environments.
    Examples:
    * Open "${homepage_url}/home"
    * Assert "id" = "sum" equals "#{5 + 6}"
    * Assert "id" = "sum" equals "#{5 + $addend}"
    In the first example, the placeholder `homepage_url` will be substituted by the environment variable with the same name.
    The substitution of placeholders is 'safe', which means, that if no variable is found, the placeholder will be unchanged.
    The second example shows a mathematical expression. Those expressions can throw exceptions, when they are invalid.
    The third example shows placeholders and mathematical expressions combined.
    Generally, placeholders are substituted first, expressions are evaluated sencond.
    """
    template = Template(gauge_param)
    #pipe operator for sets does not work on windows
    substituted = template.safe_substitute(os.environ)
    template = Template(substituted)
    substituted = template.safe_substitute(data_store.scenario)
    substituted = _substitute_expressions('#', substituted, lambda expression: array2string(numexpr.evaluate(expression)) )
    substituted = _substitute_expressions('!', substituted, lambda expression: _evaluate_expression(expression))
    return substituted


def _substitute_expressions(marker_char: str, text: str, evaluator: Callable[[str], str]) -> str:
    substituted = text
    while True:
        start = substituted.find(marker_char + '{')
        end = substituted.find('}', start)
        if start < 0 or end < 0:
            break
        expression = substituted[start + 2:end]
        before = substituted[0:start]
        after = substituted[end + 1:len(substituted)]
        value = evaluator(expression)
        substituted = before + value + after
    return substituted


def _evaluate_expression(expression: str) -> str:
    expression_lower = expression.lower()
    if expression_lower == "uuid":
        return str(uuid.uuid4())
    elif expression_lower.startswith("time"):
        if expression_lower.startswith("time:"):
            format = expression.split(':', 2)[1]
        elif expression_lower == "time":
            format = None
        else:
            raise ValueError(f"unsupported substitute {expression}")
        if format is None:
            return datetime.now().isoformat()
        else:
            return datetime.now().strftime(format)
    else:
        raise ValueError(f"unsupported substitute {expression}")



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os
import time

from getgauge.python import data_store
from selenium.webdriver import Remote
from typing import Iterable

from .app_context import app_context_key
from .config import common_config as config
from .imagepaths import ImagePath
from .images import Images


def create_screenshot(image_file_name: str) -> str:
    screenshot_file_path = _image_path().create_screenshot_file_path(image_file_name)
    _driver().save_screenshot(screenshot_file_path)
    return screenshot_file_path


def create_failure_screenshot() -> str:
    screenshot_path = _image_path().create_failure_screenshot_file_path()
    _driver().save_screenshot(screenshot_path)
    return screenshot_path


def ssim_screenshot_noscrolling(image_file_name: str, threshold: float) -> Iterable[str]:
    failed_asserts = []
    actual_screenshot_full_path = _image_path().create_actual_screenshot_file_path(image_file_name)
    _driver().save_full_page_screenshot(actual_screenshot_full_path)
    expected_screenshot_full_path = _image_path().create_expected_screenshot_file_path(image_file_name)
    if not os.path.isfile(expected_screenshot_full_path):
        failed_asserts.append("screenshot {} does not exist".format(expected_screenshot_full_path))
    else:
        append_structured_similarity(failed_asserts, expected_screenshot_full_path, actual_screenshot_full_path, threshold)
    return failed_asserts


def ssim_screenshot_scrolling(image_file_name: str, threshold: float) -> Iterable[str]:
    failed_asserts = []
    postfix = 1
    should_continue = True
    while should_continue:
        actual_screenshot_full_path = _image_path().create_actual_screenshot_file_path(image_file_name, postfix)
        _driver().save_screenshot(actual_screenshot_full_path)
        expected_screenshot_full_path = _image_path().create_expected_screenshot_file_path(image_file_name, postfix)
        should_continue = _scroll() and postfix <= 32
        postfix += 1
        if not os.path.isfile(expected_screenshot_full_path):
            failed_asserts.append("screenshot {} does not exist".format(expected_screenshot_full_path))
        else:
            append_structured_similarity(failed_asserts, expected_screenshot_full_path, actual_screenshot_full_path, threshold)
    return failed_asserts


def append_structured_similarity(asserts: list, expected_screenshot: str, actual_screenshot: str, threshold: float) -> Iterable[str]:
    diff_formats = config.get_diff_formats()
    ssim = _images().adapt_and_compare_images(expected_screenshot, actual_screenshot, diff_formats)
    if ssim < threshold:
        asserts.append("SSIM {} is less than threshold {} for {}".format(ssim, threshold, actual_screenshot))


def get_structured_similarity_to_expected(image_file_name: str, location: int, size: int, pixel_ratio: int, viewport_offset: int):
    """
    Get the structured similarity of a croped screenshot to an existing one.
    """
    actual_screenshot_full_path = _image_path().create_actual_screenshot_file_path(image_file_name)
    _driver().save_screenshot(actual_screenshot_full_path)
    crop_image(
        actual_screenshot_full_path,
        location,
        size,
        pixel_ratio,
        viewport_offset
    )
    expected_screenshot_full_path = _image_path().create_expected_screenshot_file_path(image_file_name)
    diff_formats = config.get_diff_formats()
    return _images().adapt_and_compare_images(expected_screenshot_full_path, actual_screenshot_full_path, diff_formats)


def crop_image(screenshot_path: str, location: int, size: int, pixel_ratio: int, viewport_offset: int):
    _images().crop_image_file(
        screenshot_path,
        location,
        size,
        pixel_ratio,
        viewport_offset
    )

def _scroll() -> int:
    """
    Scrolls down the size of the current window height and returns True, if scrolling down is still possible
    (The page is not ended yet)
    """
    current_offset: int = _driver().execute_script("return window.pageYOffset")
    _driver().execute_script("window.scrollBy(0, window.innerHeight)")
    wait_time = config.get_scroll_wait_time()
    time.sleep(wait_time)
    after_scroll_offset: int = _driver().execute_script("return window.pageYOffset")
    return after_scroll_offset > current_offset

def create_actual_screenshot_file_path(image_file_name: str, page: int) -> str:
    path = _image_path().create_actual_screenshot_file_path(image_file_name, page)
    _driver().save_screenshot(path)
    return path


def create_expected_screenshot_file_path(image_file_name: str, page: int) -> str:
    return _image_path().create_expected_screenshot_file_path(image_file_name, page)


def _driver() -> Remote:
    return data_store.spec[app_context_key].driver


def _image_path() -> ImagePath:
    return data_store.spec[app_context_key].image_path


def _images() -> Images:
    return data_store.spec[app_context_key].images



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import unittest
import os
import re
import time

from datetime import datetime
from getgauge.python import data_store

from gauge_web_app_steps.substitute import substitute


class TestSubstitute(unittest.TestCase):

    def setUp(self):
        self.startTime = time.time()

    def tearDown(self):
        t = time.time() - self.startTime
        print(f"{t:.3f}s")

    def test_substitute_replace(self):
        tableflip = "(ノಠ益ಠ)ノ彡┻━┻"
        os.environ["placeholder"] = tableflip
        result = substitute("${placeholder}")
        self.assertEqual(tableflip, result)

    def test_substitute_with_full_math_expression(self):
        result = substitute("#{1 + 1}")
        self.assertEqual("2", result)

    def test_substitute_with_inner_math_expression(self):
        result = substitute("\(^#{0 + 0}^)/")
        self.assertEqual("\(^0^)/", result)

    def test_substitute_with_two_math_expressions(self):
        result = substitute("#{0 + 0}#{1 + 1}")
        self.assertEqual("02", result)

    def test_substitute_complex(self):
        os.environ["a"] = "1"
        result = substitute("(${a} + 1) ** 2 = #{($a + 1) ** 2}")
        self.assertEqual("(1 + 1) ** 2 = 4", result)

    def test_substitute_fails(self):
        result = substitute("}#{1 + 1")
        self.assertEqual("}#{1 + 1", result)

    def test_substitute_with_bracket_diversions(self):
        result = substitute("}#{0 + 0}#{")
        self.assertEqual("}0#{", result)

    def test_substitute_without_pipe_operator(self):
        placeholder1 = "lala"
        placeholder2 = "baba"
        os.environ["placeholder1"] = placeholder1
        data_store.scenario["placeholder2"] = placeholder2
        result = substitute("This is ${placeholder1}/url and ${placeholder2}")
        self.assertEqual("This is lala/url and baba", result)

    def test_substitute_with_uuid_expression(self):
        result = substitute("!{uuid}")
        self.assertRegex(result, "^[0-9a-f]{8}.[0-9a-f]{4}.[0-9a-f]{4}.[0-9a-f]{4}.[0-9a-f]{12}$")

    def test_substitute_with_time_expression(self):
        result = substitute("!{time}")
        self.assertTrue(self._datetime_valid(result))

    def test_substitute_with_time_and_format_expression(self):
        result = substitute("!{time:%Y}")
        self.assertTrue(re.fullmatch("[0-9]{4}", result) is not None)

    def test_substitute_raises_with_invalid_time(self):
        self.assertRaises(ValueError, lambda: substitute("!{time-ly}"))

    def test_substitute_raises_with_invalid_expression(self):
        self.assertRaises(ValueError, lambda: substitute("!{nonexistent}"))


    def _datetime_valid(self, dt_str: str) -> bool:
        try:
            datetime.fromisoformat(dt_str)
        except ValueError:
            return False
        return True

if __name__ == '__main__':
    unittest.main()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from .browsers import Browser as Browser
from .operating_system import OperatingSystem as OperatingSystem, SaucelabsOperatingSystem as SaucelabsOperatingSystem
from .platform import Platform as Platform
from .driver_factory import DriverFactory as DriverFactory



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os
import pathlib
import unittest

from getgauge.python import ExecutionContext, Specification
from unittest.mock import patch

from gauge_web_app_steps.report import Report


class TestReport(unittest.TestCase):

    def test_spec_html_report_dir(self):
        this_directory = str(pathlib.Path().resolve())
        spec = Specification("Test Spec", os.path.join(this_directory, "resources", "test_file.spec"), False, "")
        context = ExecutionContext(spec, None, None)
        self.report = Report(context=context)
        with patch.dict(os.environ, {
            "GAUGE_PROJECT_ROOT": this_directory,
            "gauge_reports_dir": "reports"
        }):
            result = self.report._spec_html_report_dir()
            self.assertIn(os.path.join("reports", "html-report", "resources"), result)


if __name__ == '__main__':
    unittest.main()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import unittest
import os
import time

from getgauge.python import data_store
from unittest.mock import Mock, call, patch

from gauge_web_app_steps.app_context import app_context_key, timeout_key
from gauge_web_app_steps.web_app_steps import (
    answer_in_prompt,
    assert_element_does_not_exist, assert_element_exists, assert_element_is_enabled,
    before_step_hook,
    execute_async_script, execute_async_script_on_element, execute_async_script_on_element_save_result, execute_async_script_save_result,
    execute_script, execute_script_on_element, execute_script_on_element_save_result, execute_script_save_result,
    reset_timeout, save_placeholder, save_window_handles, save_window_title, set_timeout, switch_to_frame,
    wait_for_window
)


class TestWebAppSteps(unittest.TestCase):

    def setUp(self):
        self.startTime = time.time()
        data_store.scenario.clear()
        self.app_context = Mock()
        self.app_context.driver = Mock()
        self.app_context.report = Mock()
        data_store.spec[app_context_key] = self.app_context
        self.element = Mock()
        os.environ["driver_implicit_timeout"] = '0'

    def tearDown(self):
        t = time.time() - self.startTime
        print(f"{t:.3f}s")

    def test_before_step_hook(self):
        step_text = "Assert \"id\" = \"foo\" is displayed"
        new_step_text = "Assert \"id\" = \"foo\" exists"
        ctx = Mock(step = Mock(text=step_text))
        before_step_hook(ctx)
        self.app_context.report.assert_has_calls([call.log(f"The step '{step_text}' is deprecated, please use '{new_step_text}'")])

    def test_answer_in_prompt(self):
        alert = Mock()
        self.app_context.driver.switch_to.alert = alert
        answer_in_prompt("foo")
        alert.send_keys.assert_called()
        alert.accept.assert_called()
        self.app_context.driver.switch_to.default_content.assert_called()

    def test_assert_element_exists(self):
        self.element.is_displayed.return_value = True
        self.app_context.driver.find_element.return_value = self.element
        assert_element_exists("id", "foo")
        self.app_context.driver.find_element.assert_called()
        self.element.is_displayed.assert_called()

    @patch('time.sleep', return_value=None)
    def test_assert_element_exists_error(self, patched_time_sleep):
        self.element.is_displayed.return_value = False
        self.app_context.driver.find_element.return_value = self.element
        self.assertRaises(AssertionError, lambda: assert_element_exists("id", "foo"))
        self.app_context.driver.find_element.assert_called()

    def test_assert_element_does_not_exist(self):
        self.element.is_displayed.return_value = False
        self.app_context.driver.find_element.return_value = self.element
        assert_element_does_not_exist("id", "foo")
        self.app_context.driver.find_element.assert_called()
        self.element.is_displayed.assert_called()

    @patch('time.sleep', return_value=None)
    def test_assert_element_does_not_exist_error(self, patched_time_sleep):
        self.element.is_displayed.return_value = True
        self.app_context.driver.find_element.return_value = self.element
        self.assertRaises(AssertionError, lambda: assert_element_does_not_exist("id", "foo"))
        self.app_context.driver.find_element.assert_called()

    def test_assert_element_is_enabled(self):
        self.element.is_enabled.return_value = True
        self.app_context.driver.find_element.return_value = self.element
        assert_element_is_enabled("id", "foo")
        self.app_context.driver.find_element.assert_called()
        self.element.is_enabled.assert_called()

    def test_assert_element_is_enabled_error(self):
        self.element.is_enabled.return_value = False
        self.app_context.driver.find_element.return_value = self.element
        self.assertRaises(AssertionError, lambda: assert_element_is_enabled("id", "foo"))

    def test_save_placeholder(self):
        save_placeholder("placeholder-key", "placeholder_value")
        result = data_store.scenario.get("placeholder-key")
        self.assertEqual("placeholder_value", result)

    def test_save_window_handles(self):
        placeholder = "handle"
        self.app_context.driver.window_handles.return_value = ["foo"]
        save_window_handles(placeholder)
        result = data_store.scenario.get(placeholder).return_value
        self.assertEqual(["foo"], result)

    def test_save_window_title(self):
        placeholder = "title"
        self.app_context.driver.title.return_value = "foo"
        save_window_title(placeholder)
        result = data_store.scenario.get(placeholder).return_value
        self.assertEqual("foo", result)

    def test_wait_for_window(self):
        placeholder = "handle"
        self.app_context.driver.window_handles = ["foo"]
        save_window_handles(placeholder)
        self.app_context.driver.window_handles = ["foo", placeholder]
        wait_for_window("0", placeholder)
        result = data_store.scenario.get(placeholder)
        self.assertEqual(placeholder, result)

    def test_wait_for_window_missing_handles(self):
        placeholder = "handle"
        self.assertRaises(AssertionError, lambda: wait_for_window("0", placeholder))

    def test_reset_timeout(self):
        set_timeout("2")
        reset_timeout()
        self.assertIsNone(data_store.scenario.get(timeout_key))

    def test_set_timeout(self):
        set_timeout("2.5")
        self.assertEqual(2.5, data_store.scenario.get(timeout_key))

    def test_set_timeout_error(self):
        self.assertRaises(AssertionError, lambda: set_timeout("id"))

    def test_switch_to_frame_by_index(self):
        self.app_context.driver.find_elements.return_value=[]
        self.assertRaises(AssertionError, lambda: switch_to_frame("1"))
        self.app_context.driver.assert_has_calls([call.find_elements('tag name', 'frame'), call.find_elements('tag name', 'iframe')])

    def test_execute_script(self):
        execute_script("script")
        self.app_context.driver.assert_has_calls([call.execute_script("script")])

    def test_execute_script_save_result(self):
        self.app_context.driver.execute_script.return_value="result"
        execute_script_save_result("script", "placeholder")
        self.app_context.driver.assert_has_calls([call.execute_script("script")])
        self.assertEqual("result", data_store.scenario.get("placeholder"))

    def test_execute_script_on_element(self):
        elem = "web element"
        self.app_context.driver.find_element.return_value=elem
        execute_script_on_element("script(elem)", "tag name", "p", "elem")
        self.app_context.driver.assert_has_calls([
            call.find_element("tag name", "p"),
            call.execute_script("var elem=arguments[0]; script(elem)", elem)
        ])

    def test_execute_script_on_element_save_result(self):
        elem = "web element"
        self.app_context.driver.find_element.return_value=elem
        self.app_context.driver.execute_script.return_value="result"
        execute_script_on_element_save_result("script(elem)", "tag name", "p", "elem", "placeholder")
        self.app_context.driver.assert_has_calls([
            call.find_element("tag name", "p"),
            call.execute_script("var elem=arguments[0]; script(elem)", elem)
        ])
        self.assertEqual("result", data_store.scenario.get("placeholder"))

    def test_execute_async_script(self):
        execute_async_script("script")
        self.app_context.driver.assert_has_calls([call.execute_async_script("script")])

    def test_execute_async_script_save_result(self):
        self.app_context.driver.execute_async_script.return_value="result"
        execute_async_script_save_result("callback(result)", "placeholder", "callback")
        self.app_context.driver.assert_has_calls([call.execute_async_script("var callback=arguments[arguments.length-1]; callback(result)")])
        self.assertEqual("result", data_store.scenario.get("placeholder"))

    def test_execute_async_script_on_element(self):
        elem = "web element"
        self.app_context.driver.find_element.return_value=elem
        execute_async_script_on_element("script(elem)", "tag name", "p", "elem")
        self.app_context.driver.assert_has_calls([
            call.find_element("tag name", "p"),
            call.execute_async_script("var elem=arguments[0]; script(elem)", elem)
        ])

    def test_execute_async_script_on_element_save_result(self):
        elem = "web element"
        self.app_context.driver.find_element.return_value=elem
        self.app_context.driver.execute_async_script.return_value="result"
        execute_async_script_on_element_save_result("callback(elem.innerText)", "tag name", "p", "elem", "placeholder", "callback")
        self.app_context.driver.assert_has_calls([
            call.find_element("tag name", "p"),
            call.execute_async_script("var elem=arguments[0]; var callback=arguments[arguments.length-1]; callback(elem.innerText)", elem)
        ])
        self.assertEqual("result", data_store.scenario.get("placeholder"))


if __name__ == '__main__':
    unittest.main()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from getgauge.python import ExecutionContext, Specification
from selenium.webdriver import Remote

from .driver import Browser, DriverFactory
from .imagepaths import ImagePath
from .images import Images
from .report import Report
from .config import common_config as config
from .config import local_config, saucelabs_config

app_context_key = "_app_ctx"
timeout_key = "_timeout"

class AppContext:
    """
    Context objects are created and kept here.
    """

    def __init__(self, ctx: ExecutionContext = None, suite_id: str = None) -> None:
        """Initiates objects used in the basic steps."""
        if ctx is None:
            # getgauge's loading mechanism might try to instantiate the class before the lib is ready.
            return
        self.report = Report(ctx, config.is_debug_log())
        self._report_driver_options()
        spec : Specification = ctx.specification
        self.driver = self._create_driver(spec.name, suite_id)
        try:
            capabilities = self.driver.capabilities
        except AttributeError:
            capabilities = "unknown"
        self.report.log(f"Capabilities: {capabilities}")
        self.image_path = ImagePath(config.get_browser().value, config.is_headless())
        self.images = Images(self.report)
        self.diff_formats = config.get_diff_formats()
        self.mobile = config.get_operating_system().is_mobile()
        self.firefox_page_screenshot_no_scrolling = config.get_browser() == Browser.FIREFOX and config.is_whole_page_screenshot()

    def _create_driver(self, spec_name: str, suite_id: str) -> Remote:
        driver_factory = DriverFactory.create_driver_factory(spec_name, suite_id)
        return driver_factory.create_driver()

    def _report_driver_options(self) -> None:
        operating_system = config.get_operating_system()
        self.report.log(f"platform: {config.get_platform().value}")
        self.report.log(f"operating system: {operating_system}")
        self.report.log(f"operating system version: {config.get_operating_system_version()}")
        self.report.log(f"browser: {config.get_browser()}")
        self.report.log(f"page load timeout: {config.get_page_load_timeout()}")
        self.report.log(f"implicit timeout: {config.get_implicit_timeout()}")
        if config.get_platform().is_local():
            self.report.log(f"headless: {config.is_headless()}")
            if operating_system.is_mobile():
                self.report.log(f"device name: {local_config.get_mobile_device_name()}")
                self.report.log(f"real device: {local_config.get_mobile_real_device()}")
                self.report.log(f"device udid: {local_config.get_mobile_device_udid()}")
                self.report.log(f"appium server url: {local_config.get_mobile_appium_server_url()}")
        elif config.get_platform().is_remote():
            if operating_system.is_desktop():
                self.report.log(f"browser version: {saucelabs_config.get_browser_version()}")
            elif operating_system.is_mobile():
                self.report.log(f"device name: {saucelabs_config.get_device_name()}")
                self.report.log(f"appium version: {saucelabs_config.get_appium_version()}")



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from enum import Enum
from selenium.webdriver.support.select import Select


class SelectKey(Enum):
    """
    Enum for the selection type of list ements.
    """

    INDEX = "index"
    VALUE = "value"
    VISIBLE_TEXT = "visible text"

    @staticmethod
    def to_enum(key: str = ''):
        assert len(key) > 0, "expected a non empty string"
        return SelectKey[key.upper().replace(' ', '_')]


class Selector:
    """
    This class encapsulates the interaction with a select element.
    """

    @staticmethod
    def get_selected_value(element):
        return Select(element).first_selected_option.get_attribute("value")

    @staticmethod
    def select(element, select_value, select_key: SelectKey = SelectKey.INDEX):
        Selector._select(Select(element), select_value, select_key)

    @staticmethod
    def _select(select_object, select_value, select_key: SelectKey = SelectKey.INDEX):
        if select_key == SelectKey.INDEX:
            select_object.select_by_index(select_value)
        elif select_key == SelectKey.VALUE:
            select_object.select_by_value(select_value)
        elif select_key == SelectKey.VISIBLE_TEXT:
            select_object.select_by_visible_text(select_value)



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os
import requests
import subprocess
import time

from getgauge.python import data_store

from .config import saucelabs_config
from .config import common_config as config
from .driver.platform import Platform


class SauceTunnel:
    """
    This class encapsulates the interaction with the Sauce Connect Proxy.
    The tunnel can be configured with environment variables.
    Overview of configuration options: https://docs.saucelabs.com/secure-connections/sauce-connect-5/operation/configuration/
    """

    @staticmethod
    def _wait_for_sc_readiness():
        sauce_api_address = saucelabs_config.get_sauce_api_address()
        readiness_url = f"http://{sauce_api_address}/readyz"
        for _ in range(30):
            try:
                with requests.get(readiness_url) as response_readiness:
                    if response_readiness.status_code == 200 and response_readiness.text == 'OK':
                        print("Sauce Connect up")
                        time.sleep(10) # usually the readiness check is too fast
                        return
                time.sleep(2)
            except requests.RequestException:
                time.sleep(2)
        raise Exception("Sauce Connect tunnel could not be established")

    @staticmethod
    def start():
        sauce_active = saucelabs_config.is_sauce_tunnel_active()
        platform = config.get_platform()
        if not sauce_active or platform != Platform.SAUCELABS:
            return
        sauce_path = saucelabs_config.get_sauce_path()
        sc_call = f"{sauce_path} run"
        if os.name == "posix":
            sc_call = sc_call.split(" ")
        print("Starting Sauce Connect Tunnel")
        sauce_connect = subprocess.Popen(sc_call)
        data_store.suite['_sauce_connect'] = sauce_connect
        time.sleep(5)
        SauceTunnel._wait_for_sc_readiness()

    @staticmethod
    def terminate():
        sauce_connect: subprocess.Popen = data_store.suite.get('_sauce_connect')
        if sauce_connect:
            print("Terminating Sauce Connect tunnel")
            sauce_connect.terminate()
            for _ in range(10):
                time.sleep(1)
                if sauce_connect.poll() is not None:
                    return
            print("Killing Sauce Connect tunnel")
            sauce_connect.kill()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os

from typing import Optional


def get_appium_version() -> Optional[str]:
    return os.environ.get("driver_platform_saucelabs_mobile_appium_version")


def get_browser_version(default="latest") -> str:
    return os.environ.get("driver_platform_saucelabs_desktop_browser_version", default)


def get_device_name() -> Optional[str]:
    return os.environ.get("driver_platform_saucelabs_mobile_device_name")


def get_executor() -> Optional[str]:
    return os.environ.get("driver_platform_saucelabs_executor")


def get_test_title() -> Optional[str]:
    return os.environ.get("driver_platform_saucelabs_test_title")


def get_build() -> Optional[str]:
    return os.environ.get("driver_platform_saucelabs_build")


def is_device_cached() -> bool:
    return os.environ.get("driver_platform_saucelabs_devicecache", "false").lower() in ('true', '1')


def is_extended_debugging_enabled() -> bool:
    return os.environ.get("driver_platform_saucelabs_extended_debugging", "false").lower() in ('true', '1')


def is_sauce_tunnel_active() -> bool:
    return os.environ.get("SAUCE_TUNNEL_ACTIVE", "false").lower() in ('true', '1')


def get_sauce_path() -> str:
    return os.environ.get("SAUCE_PATH", "sc")


def get_sauce_user_name() -> Optional[str]:
    return os.environ.get("SAUCE_USERNAME")


def get_sauce_access_key() -> Optional[str]:
    return os.environ.get("SAUCE_ACCESS_KEY")


def get_tunnel_name() -> Optional[str]:
    return os.environ.get("SAUCE_TUNNEL_NAME")


def get_sauce_api_address() -> Optional[str]:
    return os.environ.get("SAUCE_API_ADDRESS", "127.0.0.1:8080")



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from __future__ import annotations

from enum import Enum


class OperatingSystem(Enum):

    ANDROID = "android"
    IOS = "ios"
    MACOS = "macos"
    LINUX = "linux"
    WINDOWS = "windows"

    def is_desktop(self) -> bool:
        """Checks if the given operating system a desktop system."""
        return self in [OperatingSystem.WINDOWS, OperatingSystem.MACOS, OperatingSystem.LINUX]

    def is_mobile(self) -> bool:
        """Checks if the given operating system a mobile system."""
        return self in [OperatingSystem.ANDROID, OperatingSystem.IOS]

    @staticmethod
    def parse(operating_system: str) -> OperatingSystem:
        """Try to determine an operating system from the given string."""
        operating_system = operating_system.lower()
        if operating_system in ["win", "windows"]:
            return OperatingSystem.WINDOWS
        elif operating_system in ["mac", "macos", "darwin"]:
            return OperatingSystem.MACOS
        else:
            return OperatingSystem(operating_system)

    def __str__(self) -> str:
        return self.value


class SaucelabsOperatingSystem(Enum):

    MACOS_DEFAULT = "macOS 14"
    MACOS_14 = "macOS 14"
    MACOS_13 = "macOS 13"
    MACOS_12 = "macOS 12"
    MACOS_11 = "macOS 11"
    MACOS_10_15 = "macOS 10.15"
    MACOS_10_14 = "macOS 10.14"
    MACOS_10_13 = "macOS 10.13"
    MACOS_10_12 = "macOS 10.12"
    MACOS_10_11 = "macOS 10.11"
    MACOS_10_10 = "macOS 10.10"

    WINDOWS_DEFAULT = "Windows 11"
    WINDOWS_11 = "Windows 11"
    WINDOWS_10 = "Windows 10"
    WINDOWS_8_1 = "Windows 8.1"
    WINDOWS_8 = "Windows 8"
    WINDOWS_7 = "Windows 7"

    def __str__(self) -> str:
        return self.value




#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import base64
import os
import re
import sys
import time
import traceback
import urllib
from uuid import uuid4

from itertools import filterfalse
from typing import Tuple
from getgauge.python import data_store, step, before_spec, after_spec, custom_screenshot_writer, before_suite, after_suite, before_step, ExecutionContext
from selenium.common.exceptions import JavascriptException, NoSuchWindowException, TimeoutException, WebDriverException
from selenium.webdriver import Remote
from appium.webdriver import Remote as AppiumRemote
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC

from .app_context import AppContext, app_context_key, timeout_key
from .config import common_config as config
from .driver.browsers import Browser
from .element_lookup import find_element, find_elements, find_attribute, get_text_from_element, get_marker, wait_until, wait_for_idle_element
from .keymapper import KeyMapper
from .report import Report
from .sauce_tunnel import SauceTunnel
from .selector import SelectKey, Selector
from .screenshot import (append_structured_similarity, create_screenshot, create_failure_screenshot, 
                        create_actual_screenshot_file_path, create_expected_screenshot_file_path, crop_image,
                        get_structured_similarity_to_expected, ssim_screenshot_scrolling, ssim_screenshot_noscrolling)
from .substitute import substitute


# Repeat an action a number of times before failing
max_attempts = 12
basic_auth_key = "_basic_auth"
error_message_key = "_err_msg"
suite_id_key = "_suite_id"


@before_suite
def before_suite_hook() -> None:
    try:
        SauceTunnel.start()
        data_store.suite[suite_id_key] = str(uuid4())
    except BaseException as e:
        # Gauge swallows some exceptions, so they are handled here
        print(f"An exception occured while SauceConnect initiated: {str(e)}", file=sys.stderr)
        traceback.print_exception(e)
        raise e


@after_suite
def after_suite_hook() -> None:
    try:
        SauceTunnel.terminate()
    except BaseException as e:
        # Gauge swallows some exceptions, so they are handled here
        print(f"An exception occured while SauceConnect terminated: {str(e)}", file=sys.stderr)
        traceback.print_exception(e)
        raise e


@before_spec
def before_spec_hook(exe_ctx: ExecutionContext) -> None:
    try:
        suite_id = data_store.suite[suite_id_key]
        app_ctx = AppContext(exe_ctx, suite_id)
        data_store.spec[app_context_key] = app_ctx
    except BaseException as e:
        # Gauge swallows some exceptions, so they are handled here
        print(f"An exception occured while instantiating the driver: {str(e)}", file=sys.stderr)
        traceback.print_exception(e)
        raise e


@after_spec
def after_spec_hook() -> None:
    try:
        app_ctx: AppContext = data_store.spec.get(app_context_key)
        if app_ctx is not None and app_ctx.driver is not None:
            app_ctx.report.log_debug("closing driver")
            app_ctx.driver.quit()
    except BaseException as e:
        # Gauge swallows some exceptions, so they are handled here
        print(f"An exception occured while closing the driver: {str(e)}", file=sys.stderr)
        traceback.print_exception(e)
        raise e


@before_step
def before_step_hook(exe_ctx:  ExecutionContext) -> None:
    step_text = exe_ctx.step.text
    def warn_using_deprecated_step(step_text: str, mapping: Tuple[str]) -> None:
        match = re.fullmatch(mapping[0], step_text)
        if match is not None:
            new_step = mapping[1].replace("<by>", match.group(1)).replace("<by_value>", match.group(2))
            report().log(f"The step '{step_text}' is deprecated, please use '{new_step}'")
    warn_using_deprecated_step(step_text, (r'Assert (".*?") = (".*?") is displayed', "Assert <by> = <by_value> exists",))
    warn_using_deprecated_step(step_text, (r'Assert (".*?") = (".*?") is invisible', "Assert <by> = <by_value> does not exist",))


@custom_screenshot_writer
def take_screenshot_on_failure() -> str:
    if driver() is None:
        # Error before driver initialization
        return None
    try:
        screenshot_file_path = create_failure_screenshot()
        #report().log_image(screenshot_file_path, "Step Failure Screenshot")
        return screenshot_file_path
    except Exception as e:
        report().log(str(e))
        return None


# Steps -------------------------------------------


@step("Wait <secs>")
def wait_for(secs_param: str) -> None:
    secs = substitute(secs_param)
    time.sleep(float(secs))


@step("Wait for window <secs> and save handle as <placeholder>")
def wait_for_window(secs_param: str, placeholder_name_param: str) -> None:
    secs = substitute(secs_param)
    placeholder_name = substitute(placeholder_name_param)
    assert placeholder_name in data_store.scenario.keys(), "Expected saved window handles. Did you use '* Save window handles'?"
    time.sleep(float(secs))
    wh_now = driver().window_handles
    #previous saved window handles
    wh_then = data_store.scenario[placeholder_name]
    if len(wh_now) > len(wh_then):
        win_handle = set(wh_now).difference(set(wh_then)).pop()
        data_store.scenario[placeholder_name] = win_handle


@step("Fullscreen")
def fullscreen() -> None:
    driver().fullscreen_window()
    _page_ready()


@step("Maximize")
def maximize() -> None:
    driver().maximize_window()
    _page_ready()


@step("Window size <width>x<height>")
def window_size(width_param: str, height_param: str) -> None:
    width = int(substitute(width_param))
    height = int(substitute(height_param))
    outer_width, inner_width, outer_height, inner_height = driver().execute_script("""
    return [window.outerWidth, window.innerWidth, window.outerHeight, window.innerHeight]""")
    if outer_width == 0:
        outer_width = inner_width
    if outer_height == 0:
        outer_height = inner_height
    report().log_debug("outer width: {}, inner width: {}, width: {}, outer height: {}, inner height: {}, height: {}"\
        .format(outer_width, inner_width, width, outer_height, inner_height, height))
    driver().set_window_size(outer_width - inner_width + width, outer_height - inner_height + height)
    _page_ready()


@step("Close current window")
def close_current_window() -> None:
    driver().close()
    window_handles = driver().window_handles
    last_handle = len(window_handles) - 1
    if last_handle >= 0:
        driver().switch_to.window(window_handles[last_handle])
    _page_ready()


@step("Close other windows")
def close_other_windows() -> None:
    current_handle = driver().current_window_handle
    for handle in driver().window_handles:
        if handle != current_handle:
            driver().switch_to.window(handle)
            driver().close()
    driver().switch_to.window(current_handle)
    _page_ready()


@step("Refresh")
def refresh() -> None:
    driver().refresh()
    _page_ready()


@step("Back")
def back() -> None:
    driver().back()
    _page_ready()


@step("Forward")
def forward() -> None:
    driver().forward()
    _page_ready()


@step("Open <page>")
def open_page(page_param: str) -> None:
    page = substitute(page_param)
    basic_auth_list = data_store.spec.get(basic_auth_key, [])
    webdriver = driver()
    uses_basic_auth = False
    for regexp, username, password in basic_auth_list:
        if re.match(regexp, page):
            uses_basic_auth = True
            report().log(f"URL matches `{regexp}` - using basic auth")
            if config.get_browser() in (Browser.CHROME, Browser.EDGE):
                # ToDo: rewrite this, when the BiDi API is more mature and there is a way to
                # reuse the connection over multiple Gauge steps with async/trio.
                authorization = base64.b64encode(f"{username}:{password}".encode("utf-8")).decode("ascii")
                webdriver.execute_cdp_cmd("Network.enable", {})
                webdriver.execute_cdp_cmd("Network.setExtraHTTPHeaders", ({"headers": {"Authorization": f"Basic {authorization}"}}))
                webdriver.get(page)
                webdriver.execute_cdp_cmd("Network.disable", {})
            else:
                username = username.replace('\\', '\\\\').replace("'", "\\'")
                password = password.replace('\\', '\\\\').replace("'", "\\'")
                driver().execute_script(f"""
                    const xhr = new XMLHttpRequest();
                    xhr.open('GET', '{page}', false, '{username}', '{password}');
                    xhr.send();
                """)
                webdriver.get(page)
            break
    if not uses_basic_auth:
        webdriver.get(page)
    _init_opened_page(page)


@step("Open <page> for <user>: <password>")
def open_page_for_user_and_password(page_param: str, user_param: str, password_param: str) -> None:
    report().log("This step is deprecated for security concerns. Instead use:")
    report().log("`* Register authentication <user>: <password> for <regexp>`")
    page = substitute(page_param)
    prefix_match = re.match(r"https?://", page)
    prefix = prefix_match.group(0) if prefix_match else ""
    page = page[len(prefix):]
    user = urllib.parse.quote_plus(substitute(user_param))
    password = urllib.parse.quote_plus(substitute(password_param))
    url = "{}{}:{}@{}".format(prefix, user, password, page)
    driver().get(url)
    _init_opened_page(page)


@step("Register authentication <user>: <password> for <regexp>")
def register_basic_auth_for_regexp(user_param: str, password_param: str, regexp_param: str):
    # Basic auth with CDP is not supported by all browsers yet
    username = substitute(user_param)
    password = substitute(password_param)
    regexp = substitute(regexp_param)
    auth_list = data_store.spec.get(basic_auth_key, [])
    auth_list.append((regexp, username, password,))
    data_store.spec[basic_auth_key] = auth_list


@step("Remove authentication for <regexp>")
def remove_basic_auth_for_regexp(regexp_param: str):
    regexp = substitute(regexp_param)
    auth_list = data_store.spec.get(basic_auth_key, [])
    auth_list[:] = filterfalse(lambda t: regexp == t[0], auth_list)


@step("Print window handles")
def print_window_handles() -> None:
    window_handles = driver().window_handles
    report().log("Windows: [{}]".format(", ".join(window_handles)))


@step("Print contexts")
def print_contexts() -> None:
    contexts = driver().contexts
    report().log("Contexts: [{}]".format(", ".join(contexts)))


@step("Switch to window <window_param>")
def switch_to_window(window_param: str) -> None:
    target_window = substitute(window_param)
    if target_window.isdigit():
        window_num = int(target_window)
        wait_until(lambda driver: len(driver.window_handles) > window_num, f"number of window handles is less than {window_num + 1}")
        window_name = driver().window_handles[window_num]
        driver().switch_to.window(window_name)
        report().log(f"Switched to window with index {window_num} and name {window_name}")
        _page_ready()
    else:
        original_window = driver().current_window_handle
        def find_and_switch_window(driver: Remote) -> bool:
            try:
                driver.switch_to.window(target_window)
                return True
            except NoSuchWindowException:
                for handle in driver.window_handles:
                    driver.switch_to.window(handle)
                    if driver.title == target_window:
                        return True
                return False
        try:
            wait_until(find_and_switch_window)
            _page_ready()
            report().log(f"Switched to window with name {target_window}")
        except TimeoutException:
            #switch back to default when we did not find the desired window
            driver().switch_to.window(original_window)
            raise AssertionError(f"did not find window {target_window}")


@step("Switch to default content")
def switch_to_default_content() -> None:
    driver().switch_to.default_content()


@step("Switch to frame <frame_param>")
def switch_to_frame(frame_name_or_index_param: str) -> None:
    frame_param = substitute(frame_name_or_index_param)
    if frame_param.isdigit():
        index = int(frame_param)
        frames = find_elements("tag name", "frame")
        if len(frames) == 0:
            frames = find_elements("tag name", "iframe")
        assert len(frames) > 0, "no frames or iframes found in current page."
        assert len(frames) >= index, f"frame index {index} is higher than number of frames in current page: {len(frames)}"
        driver().switch_to.frame(frames[index])
    else:
        driver().switch_to.frame(frame_param)


@step("Switch to frame <by> = <by_value>")
def switch_to_frame_by_selector(by: str, by_value: str) -> None:
    frame = find_element(by, by_value)
    driver().switch_to.frame(frame)


@step("Switch to context <context_regexp>")
def switch_to_context(context_regexp_param) -> None:
    context_regexp = substitute(context_regexp_param)
    def find_context(driver: AppiumRemote) -> str | bool:
        for c in driver.contexts:
            if re.match(context_regexp, c):
                return c
        return False
    target_context = wait_until(find_context, f"no context found that matches {context_regexp}")
    report().log(f"Switching to context '{target_context}'")
    driver().switch_to.context(target_context)


@step("Dismiss alert")
def dismiss_alert() -> None:
    wait_until(EC.alert_is_present()).dismiss()
    driver().switch_to.default_content()


@step("Accept alert")
def accept_alert() -> None:
    wait_until(EC.alert_is_present()).accept()
    driver().switch_to.default_content()


@step("Answer in prompt <text>")
def answer_in_prompt(text: str) -> None:
    prompt_text = substitute(text)
    alert = wait_until(EC.alert_is_present())
    alert.send_keys(prompt_text)
    alert.accept()
    driver().switch_to.default_content()


@step("Take a screenshot <file>")
def take_screenshot(image_file_name_param: str) -> None:
    image_file_name = substitute(image_file_name_param)
    screenshot_file_path = create_screenshot(image_file_name)
    report().log_image(screenshot_file_path)


@step("Take a screenshot of <by> = <by_value> <file>")
def take_screenshot_of_element(by: str, by_value: str, image_file_name_param: str) -> None:
    element = find_element(by, by_value)
    image_file_name = substitute(image_file_name_param)
    screenshot_file_path = create_screenshot(image_file_name)
    pixel_ratio = _device_pixel_ratio()
    viewport_offset = _viewport_offset()
    crop_image(
        screenshot_file_path,
        element.location,
        element.size,
        pixel_ratio,
        viewport_offset
    )
    report().log_image(screenshot_file_path)


@step("Take screenshots of whole page <file>")
def take_screenshots_of_whole_page(image_file_name_param: str) -> None:
    if _is_firefox_page_screenshot_no_scrolling():
        image_file_name = substitute(image_file_name_param)
        screenshot_file_path = create_screenshot(image_file_name)
        report().log_image(screenshot_file_path)
    else:
        image_file_name = substitute(image_file_name_param)
        _screenshot_of_whole_page_with_scrolling(image_file_name)


def _screenshot_of_whole_page_with_scrolling(image_file_name: str) -> None:
    postfix = 1
    should_continue = True
    while should_continue:
        filename_with_postfix = "".join((image_file_name, "_", str(postfix)))
        postfix += 1
        take_screenshot(filename_with_postfix)
        current_offset = driver().execute_script("return window.pageYOffset")
        driver().execute_script("window.scrollBy(0, window.innerHeight)")
        time.sleep(0.3)
        after_scroll_offset = driver().execute_script("return window.pageYOffset")
        if after_scroll_offset <= current_offset or postfix > 32:
            should_continue = False


@step("Check <by> = <by_value>")
def check_element(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    if not element.is_selected():
        element.click()


@step("Uncheck <by> = <by_value>")
def uncheck_element(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    if element.is_selected():
        element.click()


@step("Select <by> = <by_value> option <select_key> = <select_value>")
def select_option(by: str, by_value: str, select_key_param: str, select_value_param: str) -> None:
    element = find_element(by, by_value)
    select_key = substitute(select_key_param)
    select_value = substitute(select_value_param)
    key = SelectKey.to_enum(select_key)
    Selector.select(element, select_value, key)


@step("Click <by> = <by_value>")
def click_element(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    element.click()


@step("Double click <by> = <by_value>")
def double_click_element(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    ActionChains(driver()).double_click(element).perform()


@step("Click <by> = <by_value> <key_down>")
def click_element_with_key_down(by: str, by_value: str, key_down_param: str) -> None:
    keys_down = KeyMapper.map_keys(substitute(key_down_param))
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    ac = ActionChains(driver())
    for key in keys_down:
        ac.key_down(key)
    ac.click(element)
    for key in keys_down:
        ac.key_up(key)
    ac.perform()


@step("Right click <by> = <by_value>")
def right_click_element(by: str, by_value: str) -> None:
    actionChains = ActionChains(driver())
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    actionChains.context_click(element).perform()


@step("Type <string>")
def type_string(a_string_param: str) -> None:
    a_string = substitute(a_string_param)
    if not _is_mobile_operating_system():
        ActionChains(driver())\
            .send_keys(a_string)\
            .perform()
    else:
        element = _focused_element()
        _ready_for_interaction(element)
        element.send_keys(a_string)


@step("Type <key_down> <string>")
def type_string_with_key_down(key_down_param: str, a_string_param: str) -> None:
    keys_down = KeyMapper.map_keys(substitute(key_down_param))
    a_string = substitute(a_string_param)
    ac = ActionChains(driver())
    for key in keys_down:
        ac.key_down(key)
    ac.send_keys(a_string)
    for key in keys_down:
        ac.key_up(key)
    ac.perform()


@step("Type <by> = <by_value> <string>")
def type_string_into_element(by: str, by_value: str, a_string_param: str) -> None:
    a_string = substitute(a_string_param)
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    element.clear()
    if not _is_mobile_operating_system():
        ActionChains(driver())\
            .click(element)\
            .send_keys(a_string)\
            .perform()
    elif config.is_app_test():
        element.click()
        _focused_element().send_keys(a_string)
    else:
        element.click()
        element.send_keys(a_string)


@step("Type <by> = <by_value> <key_down> <string>")
def type_string_into_element_with_key_down(by: str, by_value: str, key_down_param: str, a_string_param: str) -> None:
    keys_down = KeyMapper.map_keys(substitute(key_down_param))
    a_string = substitute(a_string_param)
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    element.clear()
    ac = ActionChains(driver())\
        .click(element)
    for key in keys_down:
        ac.key_down(key)
    ac.send_keys(a_string)
    for key in keys_down:
        ac.key_up(key)
    ac.perform()


@step("Send keys <keys>")
def send_keys(keys_param: str) -> None:
    send_keys = KeyMapper.map_keys(substitute(keys_param))
    ActionChains(driver())\
        .send_keys(send_keys)\
        .perform()


@step("Send keys <key_down> <keys>")
def send_keys_with_key_down(key_down_param: str, keys_param: str) -> None:
    keys_down = KeyMapper.map_keys(substitute(key_down_param))
    send_keys = KeyMapper.map_keys(substitute(keys_param))
    ac = ActionChains(driver())
    for key in keys_down:
        ac.key_down(key)
    ac.send_keys(send_keys)
    for key in keys_down:
        ac.key_up(key)
    ac.perform()


@step("Send <by> = <by_value> keys <keys>")
def send_keys_to_element(by: str, by_value: str, keys_param: str) -> None:
    send_keys = KeyMapper.map_keys(substitute(keys_param))
    element = find_element(by, by_value)
    wait_until(EC.element_to_be_clickable(element))
    ActionChains(driver())\
        .click(element)\
        .send_keys(send_keys)\
        .perform()


@step("Send <by> = <by_value> keys <key_down> <keys>")
def send_keys_to_element_with_key_down(by: str, by_value: str, key_down_param: str, keys_param: str) -> None:
    keys_down = KeyMapper.map_keys(substitute(key_down_param))
    send_keys = KeyMapper.map_keys(substitute(keys_param))
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    ac = ActionChains(driver())\
        .click(element)
    for key in keys_down:
        ac.key_down(key)
    ac.send_keys(send_keys)
    for key in keys_down:
        ac.key_up(key)
    ac.perform()


@step("Clear <by> = <by_value>")
def clear_element(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    element.clear()


@step("Mouse down <by> = <by_value>")
def mouse_down(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    ActionChains(driver()).move_to_element(element).click_and_hold(element).perform()


@step("Mouse up <by> = <by_value>")
def mouse_up(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    ActionChains(driver()).move_to_element(element).release(element).perform()


@step(["Move to <by> = <by_value>", "Scroll <by> = <by_value> into view"])
def move_into_view(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    # the following script is needed for some browsers
    driver().execute_script("arguments[0].scrollIntoView(true);", element)
    try:
        ActionChains(driver()).move_to_element(element).perform()
    except WebDriverException as e:
        # in some mobile browsers it fails
        report().log_debug(f"received exception while moving to {element}: {e}")
    wait_for_idle_element(by, by_value)


@step("Move to and center <by> = <by_value>")
def move_into_view_and_center(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    driver().execute_script("arguments[0].scrollIntoView({block: 'center', inline: 'nearest'});", element)
    wait_for_idle_element(by, by_value)


@step("Move out")
def move_out_of_view() -> None:
    element = find_element("css selector", "body")
    ActionChains(driver()).move_to_element_with_offset(element, 0, 0).perform()


@step(["Move over <by> = <by_value>", "Hover over <by> = <by_value>"])
def hover_over(by: str, by_value: str) -> None:
    element = find_element(by, by_value)
    ActionChains(driver()).move_to_element(element).perform()


@step("Drag and drop <by_source> = <by_value_source> into <by_dest> = <by_value_dest>")
def dragAndDropElement(by_source: str, by_value_source: str, by_dest: str, by_value_dest: str) -> None:
    sourceElement = find_element(by_source, by_value_source)
    destinationElement = find_element(by_dest, by_value_dest)
    _ready_for_interaction(sourceElement)
    _ready_for_interaction(destinationElement)
    ActionChains(driver()).drag_and_drop(sourceElement, destinationElement).perform()


@step("Upload file = <file_path> into <by> = <by_value>")
def uploadFile(file_path_param: str, by: str, by_value: str) -> None:
    file_path = substitute(file_path_param)
    element = find_element(by, by_value)
    _ready_for_interaction(element)
    element.send_keys(file_path)


@step("Execute <script>")
def execute_script(script_param: str) -> None:
    script = substitute(script_param)
    driver().execute_script(script)


@step("Execute <script> and save result in <placeholder>")
def execute_script_save_result(script_param: str, placeholder_name_param: str) -> None:
    script = substitute(script_param)
    placeholder_name = substitute(placeholder_name_param)
    res = driver().execute_script(script)
    data_store.scenario[placeholder_name] = res


@step("Execute <script> with <by> = <by_value> as <elem>")
def execute_script_on_element(script_param: str, by: str, by_value: str, elem_param: str) -> None:
    script = substitute(script_param)
    elem = substitute(elem_param)
    assert elem in script, f"no element with name '{elem}' is referred to in the script"
    found = find_element(by, by_value)
    driver().execute_script(f"var {elem}=arguments[0]; {script}", found)


@step("Execute <script> with <by> = <by_value> as <elem> and save result in <placeholder>")
def execute_script_on_element_save_result(script_param: str, by: str, by_value: str, elem_param: str, placeholder_name_param: str) -> None:
    script = substitute(script_param)
    elem = substitute(elem_param)
    placeholder_name = substitute(placeholder_name_param)
    assert elem in script, f"no element with name '{elem}' is referred to in the script"
    found = find_element(by, by_value)
    res = driver().execute_script(f"var {elem}=arguments[0]; {script}", found)
    data_store.scenario[placeholder_name] = res


@step("Execute async <script>")
def execute_async_script(script_param: str) -> None:
    script = substitute(script_param)
    driver().execute_async_script(script)


@step("Execute async <script> and save result in <placeholder> with callback <callback>")
def execute_async_script_save_result(script_param: str, placeholder_name_param: str, callback_param: str) -> None:
    script = substitute(script_param)
    placeholder_name = substitute(placeholder_name_param)
    callback = substitute(callback_param)
    assert callback in script, f"no callback with name '{callback}' is invoked in the script"
    res = driver().execute_async_script(f"var {callback}=arguments[arguments.length-1]; {script}")
    data_store.scenario[placeholder_name] = res


@step("Execute async <script> with <by> = <by_value> as <elem>")
def execute_async_script_on_element(script_param: str, by: str, by_value: str, elem_param: str) -> None:
    script = substitute(script_param)
    elem = substitute(elem_param)
    assert elem in script, f"no element with name '{elem}' is referred to in the script"
    found = find_element(by, by_value)
    driver().execute_async_script(f"var {elem}=arguments[0]; {script}", found)


@step("Execute async <script> with <by> = <by_value> as <elem> and save result in <placeholder> with callback <callback>")
def execute_async_script_on_element_save_result(
    script_param: str,
    by: str,
    by_value: str,
    elem_param: str,
    placeholder_name_param: str,
    callback_param: str
) -> None:
    script = substitute(script_param)
    elem = substitute(elem_param)
    placeholder_name = substitute(placeholder_name_param)
    callback = substitute(callback_param)
    assert elem in script, f"no element with name '{elem}' is referred to in the script"
    assert callback in script, f"no callback with name '{callback}' is invoked in the script"
    found = find_element(by, by_value)
    res = driver().execute_async_script(f"var {elem}=arguments[0]; var {callback}=arguments[arguments.length-1]; {script}", found)
    data_store.scenario[placeholder_name] = res


@step("Save placeholder <placeholder> = <value>")
def save_placeholder(placeholder_name_param: str, placeholder_value_param: str) -> None:
    placeholder_name = substitute(placeholder_name_param)
    placeholder_value = substitute(placeholder_value_param)
    data_store.scenario[placeholder_name] = placeholder_value


@step("Save placeholder <placeholder> from <by> = <by_value>")
def save_placeholder_from_element(placeholder_name_param: str, by: str, by_value: str) -> None:
    placeholder_name = substitute(placeholder_name_param)
    text = get_text_from_element(by, by_value)
    data_store.scenario[placeholder_name] = text


@step("Save placeholder <placeholder> from attribute <attribute_param> of <by> = <by_value>")
def save_placeholder_from_element_attribute(placeholder_name_param: str, attribute_param: str, by: str, by_value: str) -> None:
    placeholder_name = substitute(placeholder_name_param)
    attribute = substitute(attribute_param)
    attribute_value = find_attribute(by, by_value, attribute)
    data_store.scenario[placeholder_name] = attribute_value


@step("Save window handles as <placeholder>")
def save_window_handles(placeholder_name_param: str) -> None:
    placeholder_name = substitute(placeholder_name_param)
    data_store.scenario[placeholder_name] = driver().window_handles


@step("Save window title as <placeholder>")
def save_window_title(placeholder_name_param: str) -> None:
    placeholder_name = substitute(placeholder_name_param)
    data_store.scenario[placeholder_name] = driver().title


@step("Set timeout <seconds>")
def set_timeout(seconds_param: str):
    seconds = substitute(seconds_param)
    assert seconds.replace('.', '', 1).isdigit(),\
        _err_msg(f"argument '{seconds_param}' should be a number")
    data_store.scenario[timeout_key] = float(seconds)


@step("Reset timeout")
def reset_timeout():
    del data_store.scenario[timeout_key]


# Steps Asserts ------------------------------------------------


@step("Show message in an error case <error_message>")
def show_error_message_in_case(error_msg_param: str) -> None:
    error_msg = substitute(error_msg_param)
    data_store.scenario[error_message_key] = error_msg


@step("Assert window handles is <windows_num>")
def assert_window_handle_num(window_num_param: str) -> None:
    window_num = substitute(window_num_param)
    expected = int(window_num)
    try:
        wait_until(lambda driver: expected == len(driver.window_handles))
    except TimeoutException:
        actual = len(driver().window_handles)
        raise AssertionError(_err_msg(f"expected {expected} window handles, got {actual}"))


@step("Assert title equals <title>")
def assert_title(expected_title_param: str) -> None:
    expected_title = substitute(expected_title_param)
    try:
        wait_until(lambda driver: driver.title == expected_title)
    except TimeoutException:
        current_title = driver().title
        raise AssertionError(_err_msg(f"expected title: {expected_title}, actual {current_title}"))


@step("Assert dialog text equals <text>")
def assert_dialog_text(expected_text_param: str) -> None:
    expected_text = substitute(expected_text_param)
    try:
        wait_until(lambda driver: driver.switch_to.alert.text == expected_text)
    except TimeoutException:
        dialog_text = driver().switch_to.alert.text
        # if no alert is present, a NoAlertPresentException is raised, which is fine
        raise AssertionError(_err_msg(f"expected dialog text: {expected_text}, actual {dialog_text}"))


@step("Assert url equals <url>")
def assert_url(expected_url_param: str) -> None:
    expected_url = substitute(expected_url_param)
    try:
        wait_until(lambda driver: driver.current_url == expected_url)
    except TimeoutException:
        current_url = driver().current_url
        raise AssertionError(_err_msg(f"url does not equal {expected_url}, got {current_url}"))


@step("Assert url starts with <url>")
def assert_url_starts_with(expected_url_param: str) -> None:
    expected_url = substitute(expected_url_param)
    try:
        wait_until(lambda driver: driver.current_url.startswith(expected_url))
    except TimeoutException:
        current_url = driver().current_url
        raise AssertionError(_err_msg(f"url does not start with {expected_url}, got {current_url}"))


@step("Assert url ends with <url>")
def assert_url_ends_with(expected_url_param: str) -> None:
    expected_url = substitute(expected_url_param)
    try:
        wait_until(lambda driver: driver.current_url.endswith(expected_url))
    except TimeoutException:
        current_url = driver().current_url
        raise AssertionError(_err_msg(f"url does not end with {expected_url}, got {current_url}"))


@step("Assert url contains <url>")
def assert_url_contains(expected_url_param: str) -> None:
    expected_url = substitute(expected_url_param)
    try:
        wait_until(lambda driver: expected_url in driver.current_url)
    except TimeoutException:
        current_url = driver().current_url
        raise AssertionError(_err_msg(f"url does not contain {expected_url}, got {current_url}"))


# see also before_step_hook
@step(["Assert <by> = <by_value> exists", "Assert <by> = <by_value> is displayed"])
def assert_element_exists(by: str, by_value: str) -> None:
    marker = get_marker(substitute(by), substitute(by_value))
    try:
        wait_until(EC.visibility_of_element_located(marker))
    except TimeoutException:
        raise AssertionError(_err_msg(f"element {by} = {by_value} does not exist"))


# see also before_step_hook
@step(["Assert <by> = <by_value> does not exist", "Assert <by> = <by_value> is invisible"])
def assert_element_does_not_exist(by_param: str, by_value_param: str) -> None:
    by = substitute(by_param)
    by_value = substitute(by_value_param)
    marker = get_marker(by, by_value)
    try:
        wait_until(EC.invisibility_of_element(marker))
    except TimeoutException:
        raise AssertionError(_err_msg(f"element {by} = {by_value} exists"))


@step("Assert <by> = <by_value> is enabled")
def assert_element_is_enabled(by: str, by_value: str) -> None:
    try:
        wait_until(lambda _: find_element(by, by_value).is_enabled())
    except TimeoutException:
        raise AssertionError(_err_msg(f"element {by} = {by_value} is disabled"))


@step("Assert <by> = <by_value> is disabled")
def assert_element_is_disabled(by: str, by_value: str) -> None:
    try:
        wait_until(lambda _: not find_element(by, by_value).is_enabled())
    except TimeoutException:
        raise AssertionError(_err_msg(f"element {by} = {by_value} is enabled"))


@step("Assert <by> = <by_value> is selected")
def assert_element_is_selected(by: str, by_value: str) -> None:
    try:
        wait_until(lambda _: find_element(by, by_value).is_selected())
    except TimeoutException:
        element = find_element(by, by_value, immediately=True)
        if element is None:
            raise AssertionError(_err_msg(f"no element {by} = {by_value} found"))
        raise AssertionError(_err_msg(f"element {by} = {by_value} is not selected"))


@step("Assert <by> = <by_value> has selected value <value>")
def assert_selected_option(by: str, by_value: str, expected_param: str) -> None:
    expected = substitute(expected_param)
    try:
        wait_until(lambda _: Selector.get_selected_value(find_element(by, by_value)) == expected)
    except TimeoutException:
        element = find_element(by, by_value, immediately=True)
        if element is None:
            raise AssertionError(_err_msg(f"no element {by} = {by_value} found"))
        actual = Selector.get_selected_value(element)
        raise AssertionError(_err_msg(f"expected value: {expected}, actual {actual}"))


@step("Assert <by> = <by_value> is not selected")
def assert_element_is_not_selected(by: str, by_value: str) -> None:
    try:
        wait_until(lambda _: not find_element(by, by_value).is_selected())
    except TimeoutException:
        element = find_element(by, by_value, immediately=True)
        if element is None:
            raise AssertionError(_err_msg(f"no element {by} = {by_value} found"))
        raise AssertionError(_err_msg(f"element {by} = {by_value} is selected"))


@step("Assert <by> = <by_value> equals <string>")
def assert_text_equals(by: str, by_value: str, expected_text_param: str) -> None:
    expected_text = substitute(expected_text_param)
    try:
        wait_until(lambda _: get_text_from_element(by, by_value) == expected_text)
    except TimeoutException:
        text = get_text_from_element(by, by_value, immediately=True)
        raise AssertionError(_err_msg(f"element {by} = {by_value} expected text {expected_text}, actual {text}"))


@step("Assert <by> = <by_value> does not equal <string>")
def assert_text_does_not_equal(by: str, by_value: str, expected_text_param: str) -> None:
    expected_text = substitute(expected_text_param)
    try:
        wait_until(lambda _: get_text_from_element(by, by_value) != expected_text)
    except TimeoutException:
        text = get_text_from_element(by, by_value, immediately=True)
        raise AssertionError(_err_msg(f"element {by} = {by_value} has unexpected text {text}"))


@step("Assert <by> = <by_value> regexp <regexp>")
def assert_regexp_in_element(by: str, by_value: str, regexp_param: str) -> None:
    regexp = substitute(regexp_param)
    try:
        wait_until(lambda _: re.match(regexp, get_text_from_element(by, by_value)))
    except TimeoutException:
        text = get_text_from_element(by, by_value, immediately=True)
        raise AssertionError(_err_msg(f"element {by} = {by_value}: regexp {regexp} does not match {text}"))


@step("Assert <by> = <by_value> contains <string>")
def assert_text_contains(by: str, by_value: str, contains_text_param: str) -> None:
    contains_text = substitute(contains_text_param)
    try:
        wait_until(lambda _: contains_text in get_text_from_element(by, by_value))
    except TimeoutException:
        text = get_text_from_element(by, by_value, immediately=True)
        raise AssertionError(_err_msg(f"element {by} = {by_value} expected actual {text} to contain {contains_text}"))


@step("Assert <by> = <by_value> does not contain <string>")
def assert_text_does_not_contain(by: str, by_value: str, contains_text_param: str) -> None:
    contains_text = substitute(contains_text_param)
    try:
        wait_until(lambda _: contains_text not in get_text_from_element(by, by_value))
    except TimeoutException:
        text = get_text_from_element(by, by_value, immediately=True)
        raise AssertionError(_err_msg(f"element {by} = {by_value} expected actual {text} to not contain {contains_text}"))


@step("Assert <by> = <by_value> css <css_property_name> is <css_expected_value>")
def assert_css_property(by: str, by_value: str, css_property_name_param: str, css_expected_value_param: str) -> None:
    css_property_name = substitute(css_property_name_param)
    css_expected_value = substitute(css_expected_value_param)
    try:
        wait_until(lambda _: css_expected_value == find_element(by, by_value).value_of_css_property(css_property_name))
    except TimeoutException:
        element = find_element(by, by_value, immediately=True)
        css_actual_value = element.value_of_css_property(css_property_name) if element is not None else "None"
        raise AssertionError(_err_msg(f"element {by} = {by_value}: css property {css_property_name} expected: {css_expected_value}, actual: {css_actual_value}"))


@step("Assert <by> = <by_value> is focused")
def assert_element_is_focused(by: str, by_value: str) -> None:
    try:
        wait_until(lambda _: find_element(by, by_value) == _focused_element())
    except TimeoutException:
        raise AssertionError(_err_msg(f"element {by} = {by_value} is not in focus"))


@step("Assert <by> = <by_value> attribute <attribute> exists")
def assert_attribute_exists(by: str, by_value: str, attribute_param: str) -> None:
    attribute = substitute(attribute_param)
    try:
        wait_until(lambda _: find_attribute(by, by_value, attribute))
    except TimeoutException:
        raise AssertionError(_err_msg(f"element {by} = {by_value} has no attribute {attribute}"))


@step("Assert <by> = <by_value> attribute <attribute> contains <value>")
def assert_attribute_contains(by: str, by_value: str, attribute_param: str, value_param: str) -> None:
    attribute = substitute(attribute_param)
    value = substitute(value_param)
    def attribute_contains_value(_: Remote) -> bool:
        found = find_attribute(by, by_value, attribute)
        return isinstance(found, str) and value in found
    try:
        wait_until(attribute_contains_value)
    except TimeoutException:
        found_value = find_attribute(by, by_value, attribute, immediately=True)
        if found_value is None:
            msg = _err_msg(f"element {by} = {by_value} has no attribute {attribute}")
        elif found_value is True:
            msg = _err_msg(f"attribute {attribute} in element {by} = {by_value} is empty")
        else:
            msg = _err_msg(f"attribute {attribute} in element {by} = {by_value} does not contain {value} - found: {found_value}")
        raise AssertionError(msg)


@step("Assert <by> = <by_value> attribute <attribute> equals <value>")
def assert_attribute_equals(by: str, by_value: str, attribute_param: str, value_param: str) -> None:
    attribute = substitute(attribute_param)
    value = substitute(value_param)
    try:
        wait_until(lambda _: value == find_attribute(by, by_value, attribute))
    except TimeoutException:
        found_value = find_attribute(by, by_value, attribute, immediately=True)
        if found_value is None:
            msg = _err_msg(f"element {by} = {by_value} has no attribute {attribute}")
        elif found_value is True:
            msg = _err_msg(f"attribute {attribute} in element {by} = {by_value} is empty")
        else:
            msg = _err_msg(f"attribute {attribute} in element {by} = {by_value} does not equal {value} - found: {found_value}")
        raise AssertionError(msg)


@step("Assert <by> = <by_value> attribute <attribute> does not contain <value>")
def assert_attribute_does_not_contain(by: str, by_value: str, attribute_param: str, value_param: str) -> None:
    attribute = substitute(attribute_param)
    value = substitute(value_param)
    def attribute_does_not_contain_value(_: Remote) -> bool:
        found = find_attribute(by, by_value, attribute)
        return isinstance(found, str) and value not in found
    try:
        wait_until(attribute_does_not_contain_value)
    except TimeoutException:
        found_value = find_attribute(by, by_value, attribute, immediately=True)
        raise AssertionError(_err_msg(f"attribute {attribute} in element {by} = {by_value} contains {value} - found: {found_value}"))


@step("Assert <by> = <by_value> screenshot resembles <file> with SSIM more than <threshold>")
def assert_image_resembles(by: str, by_value: str, image_file_name_param: str, threshold_param: str) -> None:
    element = find_element(by, by_value)
    threshold = float(substitute(threshold_param))
    assert 0.0 <= threshold <= 1.0, "threshold must be between 0.0 and 1.0"
    pixel_ratio = _device_pixel_ratio()
    viewport_offset = _viewport_offset()
    image_file_name = substitute(image_file_name_param)
    ssim = get_structured_similarity_to_expected(image_file_name, element.location, element.size, pixel_ratio, viewport_offset)
    assert ssim >= threshold, \
        _err_msg(f"SSIM {ssim} is less than threshold {threshold}")


@step("Assert page screenshots resemble <file> with SSIM more than <threshold>")
def assert_whole_page_resembles(image_file_name_param: str, threshold_param: str) -> None:
    threshold = float(substitute(threshold_param))
    assert 0.0 <= threshold <= 1.0, "threshold must be between 0.0 and 1.0"
    image_file_name = substitute(image_file_name_param)
    if _is_firefox_page_screenshot_no_scrolling():
        failed_asserts = ssim_screenshot_noscrolling(image_file_name, threshold)
    else:
        failed_asserts = ssim_screenshot_scrolling(image_file_name, threshold)
    assert len(failed_asserts) == 0,\
            _err_msg("Assertions failed:\n\t{}".format("\n\t".join(failed_asserts)))


@step("Assert page screenshots resemble <file> with SSIM more than <threshold> for <pages> pages")
def assert_pages_resemble(image_file_name_param: str, threshold_param: str, pages_param: str) -> None:
    """This step exists in case the upper step does not work, f.i. due to framework restrictions"""
    threshold = float(substitute(threshold_param))
    assert 0.0 <= threshold <= 1.0, "threshold must be between 0.0 and 1.0"
    image_file_name = substitute(image_file_name_param)
    pages = int(substitute(pages_param))
    failed_asserts = []
    for page in range(1, pages + 1):
        actual_screenshot_full_path = create_actual_screenshot_file_path(image_file_name, page)
        expected_screenshot_full_path = create_expected_screenshot_file_path(image_file_name, page)
        page += 1
        send_keys("PAGE_DOWN")
        wait_time = config.get_scroll_wait_time()
        time.sleep(wait_time)
        if not os.path.isfile(expected_screenshot_full_path):
            failed_asserts.append("screenshot {} does not exist".format(expected_screenshot_full_path))
        else:
            append_structured_similarity(failed_asserts, expected_screenshot_full_path, actual_screenshot_full_path, threshold)
    assert len(failed_asserts) == 0,\
            _err_msg("Assertions failed:\n\t{}".format("\n\t".join(failed_asserts)))


@step("Fail <message>")
def fail(message_param: str) -> None:
    message = substitute(message_param)
    report().log(message)
    assert False

@step("Print message <message>")
def print_message(message_param: str) -> None:
    message = substitute(message_param)
    report().log(message)

# Context methods -------------------------------------------


def driver() -> Remote | AppiumRemote:
    app_ctx: AppContext = data_store.spec[app_context_key]
    return app_ctx.driver


def report() -> Report:
    app_ctx: AppContext = data_store.spec[app_context_key]
    return app_ctx.report


# Private methods -------------------------------------------

def _init_opened_page(page) -> None:
    for _ in range(max_attempts):
        # Browsers sometimes open the page in the middle, when visited before
        _page_ready()
        if "#" not in page:
            try:
                driver().execute_script("window.scrollTo(0, 0);")
                break
            except JavascriptException:
                # In case of a page redirect the window object can become invalid.
                pass


def _page_ready() -> None:
    if config.is_app_test():
        # cannot execute script in native apps
        return
    try:
        wait_until(lambda driver: "complete" == driver.execute_script("return document.readyState"))
    except TimeoutException:
        # Some dynamic pages will never be ready
        pass


def _ready_for_interaction(element: WebElement) -> None:
    try:
        wait_until(EC.element_to_be_clickable(element))
    except TimeoutException:
        # best chance. But false negatives possible.
        pass


def _device_pixel_ratio() -> int:
    return int(driver().execute_script("return window.devicePixelRatio"))


def _viewport_offset() -> int:
    return int(driver().execute_script("return window.pageYOffset"))

def _is_firefox_page_screenshot_no_scrolling() -> bool:
    return config.get_browser() == Browser.FIREFOX and config.is_whole_page_screenshot()


def _is_mobile_operating_system() -> bool:
    return config.get_operating_system().is_mobile()


def _focused_element() -> WebElement:
    return driver().switch_to.active_element


def _err_msg(default_msg: str) -> str:
    return str(data_store.scenario.get(error_message_key, default_msg))



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from enum import Enum


class Platform(Enum):

    LOCAL = "local"
    SAUCELABS = "saucelabs"

    def is_local(self) -> bool:
        return self == Platform.LOCAL

    def is_remote(self) -> bool:
        return self == Platform.SAUCELABS

    def __str__(self) -> str:
        return self.value




#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os
import glob
import shutil
import unittest

from getgauge.python import data_store
from pathlib import Path
from unittest.mock import MagicMock, patch

from gauge_web_app_steps.web_app_steps import (
    suite_id_key,
    after_spec_hook, accept_alert, answer_in_prompt,
    assert_attribute_contains, assert_attribute_does_not_contain, assert_attribute_equals, assert_attribute_exists,
    assert_dialog_text, assert_element_exists, assert_element_does_not_exist,
    assert_element_is_selected, assert_element_is_not_selected, assert_selected_option,
    assert_text_contains, assert_text_does_not_contain, assert_text_does_not_equal, assert_text_equals,
    assert_title, assert_whole_page_resembles,
    before_spec_hook, check_element, click_element, double_click_element, driver,
    hover_over, move_into_view, move_out_of_view, mouse_down, mouse_up, open_page,
    save_placeholder_from_element, save_placeholder_from_element_attribute,
    select_option, switch_to_frame, switch_to_window,
    take_screenshots_of_whole_page, type_string, uncheck_element, wait_for

)
from gauge_web_app_steps.driver import Browser


@unittest.skipIf(os.environ.get("TEST_SKIP_IT", "0") == "1", "TEST_SKIP_IT env variable is set")
class TestWebAppStepsIT(unittest.TestCase):

    MAIN_PAGE: str = "https://the-internet.herokuapp.com"

    @classmethod
    def setUpClass(cls):
        data_store.suite.clear()
        data_store.suite[suite_id_key] = "suite-id"
        test_dir = os.path.abspath(os.path.dirname(__file__))
        cls.out = os.path.join(test_dir, "out")
        cls.browser = Browser.CHROME.value
        cls.env_patcher = patch.dict(os.environ, {
            "GAUGE_PROJECT_ROOT": cls.out,
            "driver_browser": cls.browser,
            "driver_platform_local_headless": "False",
            "driver_implicit_timeout": "7",
        })
        cls.env_patcher.start()
        before_spec_hook(MagicMock())

    @classmethod
    def tearDownClass(cls):
        after_spec_hook()
        cls.env_patcher.stop()


    def setUp(self) -> None:
        self.screenshots_dir = os.path.join(TestWebAppStepsIT.out, "screenshots")

    def tearDown(self):
        if os.path.exists(self.screenshots_dir):
            shutil.rmtree(self.screenshots_dir)

    def _files(self):
        return glob.glob(self.screenshots_dir + f"{os.path.sep}*")

    def _open_page(self, page=MAIN_PAGE, sub_page=""):
        open_page(page + sub_page)
        wait_for("1")

    def test_assert_attribute(self):
        self._open_page()
        assert_attribute_contains("css selector", "#content", "class", "large-12")
        assert_attribute_equals("css selector", "#content", "class", "large-12 columns")
        assert_attribute_does_not_contain("css selector", "#content", "class", "large-13")

    def test_assert_attribute_exists(self):
        self._open_page(sub_page="/checkboxes")
        assert_attribute_exists("css selector", "#checkboxes input[checked]", "checked")

    def test_answer_in_prompt(self):
        self._open_page(sub_page="/javascript_alerts")
        click_element("xpath", "(//button)[3]")
        answer_in_prompt("foo")
        assert_text_equals("xpath", "//*[@id='result']", "You entered: foo")

    def test_assert_dialog_text(self):
        self._open_page(sub_page="/javascript_alerts")
        click_element("xpath", "(//button)[1]")
        assert_dialog_text("I am a JS Alert")
        accept_alert()

    def test_assert_dynamic_element_displayed(self):
        self._open_page(sub_page="/dynamic_loading/1")
        assert_element_does_not_exist("id", "finish")
        click_element("xpath", "//button")
        assert_element_exists("id", "finish")

    def test_assert_text(self):
        self._open_page()
        assert_text_contains("tag name", "h2", "Examples")
        assert_text_does_not_contain("tag name", "h2", "Shmexamples")

    def test_check(self):
        self._open_page(sub_page="/checkboxes")
        check_element("xpath", "(//input)[1]")
        uncheck_element("xpath", "(//input)[2]")
        assert_element_is_selected("xpath", "(//input)[1]")
        assert_element_is_not_selected("xpath", "(//input)[2]")

    def test_double_click(self):
        self._open_page(sub_page="/add_remove_elements/")
        double_click_element("xpath", "(//button)[1]")
        assert_text_equals("xpath", "(//button)[2]", "Delete")
        assert_text_does_not_equal("xpath", "(//button)[2]", "delete")
        assert_text_equals("xpath", "(//button)[3]", "Delete")

    def test_hover_over(self):
        self._open_page(sub_page="/hovers")
        hover_over("css selector", "div.figure:nth-child(3)")
        assert_element_exists("css selector", "div.figure:nth-child(3) > div.figcaption")

    def test_mouse_down_and_release(self):
        self._open_page(sub_page="/add_remove_elements/")
        mouse_down("xpath", "(//button)[1]")
        mouse_up("xpath", "(//button)[1]")
        assert_element_exists("xpath", "(//button)[2]")

    def test_move(self):
        self._open_page(sub_page="/hovers")
        move_into_view("css selector", "div.figure:nth-child(3)")
        assert_element_exists("css selector", "div.figure:nth-child(3) > div.figcaption")
        move_out_of_view()
        assert_element_does_not_exist("css selector", "div.figure:nth-child(3) > div.figcaption")

    def test_save_placeholder_from_element(self):
        self._open_page(sub_page="/inputs")
        click_element("xpath", "//input")
        type_string("5")
        save_placeholder_from_element("mynum", "tag name", "input")
        result = data_store.scenario.get("mynum")
        self.assertEqual("5", result)

    def test_save_placeholder_from_element_attribute(self):
        self._open_page()
        save_placeholder_from_element_attribute("myref", "href", "xpath", "//li[1]/a")
        result = data_store.scenario.get("myref")
        self.assertEqual("/abtest", result)

    def test_select_list_element(self):
        self._open_page(page="https://mdn.github.io/html-examples/custom-select/")
        select_option("id", "select", "index", "2")
        assert_selected_option("id", "select", "Beans")

    def test_switch_to_frame_by_name(self):
        self._open_page(sub_page="/nested_frames")
        switch_to_frame("frame-bottom")
        assert_text_equals("xpath", "/html/body", "BOTTOM")

    def test_switch_to_frame_by_index(self):
        self._open_page(sub_page="/nested_frames")
        switch_to_frame("1")
        assert_text_equals("xpath", "/html/body", "BOTTOM")

    def test_take_screenshots_of_whole_page_scrolling(self):
        os.environ["screenshot_whole_page_no_scroll"] = "False"
        self._open_page()
        take_screenshots_of_whole_page(self.__class__.__name__)
        self.assertGreater(len(self._files()), 1)

    #works only with Firefox => TODO: skip if not Firefox
    @unittest.SkipTest
    def test_take_screenshots_of_whole_page_no_scrolling(self):
        os.environ["screenshot_whole_page_no_scroll"] = "True"
        self._open_page()
        take_screenshots_of_whole_page(self.__class__.__name__)
        file = f"{TestWebAppStepsIT.browser}_{self.__class__.__name__}.png"
        src = os.path.join(self.screenshots_dir, file)
        dst_dir = os.path.join(self.out, "expected_screenshots")
        dst = os.path.join(dst_dir, file)
        Path(dst_dir).mkdir(exist_ok=True)
        shutil.copy(src, dst)
        assert_whole_page_resembles(self.__class__.__name__, "1.0")
        self.assertEqual(len(self._files()), 1)

    #the following tests should be executed at last because the switch windows
    def test_window_by_index(self):
        self._open_page()
        driver().switch_to.new_window('window')
        switch_to_window("0")
        assert_title("The Internet")

    def test_window_by_name(self):
        self._open_page()
        driver().switch_to.new_window('window')
        switch_to_window("The Internet")
        assert_title("The Internet")


if __name__ == '__main__':
    unittest.main()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from appium.webdriver.common.appiumby import AppiumBy
from selenium.webdriver.common.by import By


class ByMapper:

    _MAPPING = {
        "accessibility id": AppiumBy.ACCESSIBILITY_ID,
        "id": By.ID,
        "xpath": By.XPATH,
        "link text": By.LINK_TEXT,
        "partial link text": By.PARTIAL_LINK_TEXT,
        "name": By.NAME,
        "tag name": By.TAG_NAME,
        "class name": By.CLASS_NAME,
        "css selector": By.CSS_SELECTOR,
    }

    @classmethod
    def map_string(cls, by_string: str) -> str:
        return cls._MAPPING[by_string]



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os

from typing import Optional


def get_mobile_appium_server_url() -> Optional[str]:
    return os.environ.get("driver_platform_local_mobile_appium_server_url")


def get_mobile_device_name() -> Optional[str]:
    return os.environ.get("driver_platform_local_mobile_device_name")


def get_mobile_device_udid() -> Optional[str]:
    return os.environ.get("driver_platform_local_mobile_device_udid")


def get_mobile_real_device(default="false") -> bool:
    return os.environ.get("driver_platform_local_mobile_real_device", default).lower() in ("true", "1")



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os

from collections.abc import Iterable
from getgauge.python import Messages, ExecutionContext


class Report(object):
    """
    Here you find methods related to writing to gauge reports.
    """

    def __init__(self, context: ExecutionContext = None, debug=False) -> None:
        self.context = context
        self.debug = debug

    def _spec_html_report_dir(self) -> str:
        gauge_project_dir = os.environ.get("GAUGE_PROJECT_ROOT")
        spec_file = self.context.specification.file_name
        spec_dir = os.path.dirname(spec_file)
        spec_dir_rel = spec_dir[len(gauge_project_dir)+1:]
        gauge_reports_dir = os.environ.get("gauge_reports_dir", "reports")
        html_report_dir = os.path.join(gauge_reports_dir, "html-report")
        if not os.path.isabs(html_report_dir):
            html_report_dir = os.path.join(gauge_project_dir, html_report_dir)
        spec_report_dir = os.path.join(html_report_dir, spec_dir_rel)
        return spec_report_dir

    def log(self, message="") -> None:
        """
        Log to terminal and write into the gauge report the given string
        """
        print(message)  # > terminal
        Messages.write_message(message)  # > html report

    def log_debug(self, message="") -> None:
        if not self.debug:
            return
        self.log(message)

    def log_image(self, image_file_path, label="") -> None:
        """
        Include an image into the gauge report.
        The relative path from the report to the image will be used in the created html.
        """
        spec_report_dir = self._spec_html_report_dir()
        rel_path_report_to_image = os.path.relpath(image_file_path, spec_report_dir)
        html_rel_path = rel_path_report_to_image.replace('\\', '/')
        Messages.write_message(
            "<a href='{0}'><img src='{0}' height='165'/></a><label>{1}</label>".format(html_rel_path, label))

    def log_image_info(self, name, image) -> None:
        """
        Include information about the given image into the gauge report.
        The image parameter is expected to be a multidimensional array,
        just as the skimage library works with.
        """
        if not self.debug:
            return
        self.log()
        self.log("Image Info: %s" % (name,))
        self.log("Size: %sx%s" % (len(image[0]), len(image),))
        depth = 1
        channel = image
        while isinstance(channel, Iterable):
            self.log("Channel %s length: %s" % (depth, len(channel),))
            if not isinstance(channel[0], Iterable):
                self.log("Last channel values excerpt: [%s]" % (", ".join([str(i) for i in channel]),))
            channel = channel[0]
            depth += 1
        self.log("Channels: %s" % (depth,))
        self.log()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import numpy as np
import os
import shutil
import unittest
from numpy import uint8
from skimage import img_as_ubyte, io
from unittest.mock import MagicMock
from parameterized import parameterized

from gauge_web_app_steps.images import Images
from tests import TEST_RESOURCES_DIR, TEST_OUT_DIR


class TestImages(unittest.TestCase):

    def setUp(self) -> None:
        self.test_instance = Images(MagicMock())
        self.actual_image = os.path.join(TEST_OUT_DIR, "actual_screenshots", "actual_rgba.png")
        self.actual_image_rgb = os.path.join(TEST_OUT_DIR, "actual_screenshots", "actual_rgb.png")
        self.expected_image = os.path.join(TEST_RESOURCES_DIR, "expected_rgba.png")
        self.expected_image_rgb = os.path.join(TEST_RESOURCES_DIR, "expected_rgb.png")
        self.diffs_dir = os.path.join(TEST_OUT_DIR, "diffs")
        self.crop_dir = os.path.join(TEST_OUT_DIR, "crop")
        os.makedirs(os.path.join(TEST_OUT_DIR, "actual_screenshots"), exist_ok=True)
        os.makedirs(self.diffs_dir, exist_ok=True)
        os.makedirs(self.crop_dir, exist_ok=True)
        # the actual image file might be altered during the test, so the test runs on a copy
        shutil.copy(os.path.join(TEST_RESOURCES_DIR, "actual_rgb.png"), self.actual_image_rgb)
        shutil.copy(os.path.join(TEST_RESOURCES_DIR, "actual_rgba.png"), self.actual_image)

    @parameterized.expand(["full", "gradient", "red", "color:fuchsia"])
    def test_adapt_and_compare_images(self, diff_format:str):
        mergefile = os.path.join(self.diffs_dir, "actual_rgba_merged.png")
        self._remove_image_if_it_exists(mergefile)
        expected_diff_file = os.path.join(self.diffs_dir, f"actual_rgba_{diff_format.removeprefix('color:')}.png")
        self._remove_image_if_it_exists(expected_diff_file)
        ssim = self.test_instance.adapt_and_compare_images(
            expected_screenshot_full_path=self.expected_image,
            actual_screenshot_full_path=self.actual_image,
            output_path=self.diffs_dir,
            diff_formats=diff_format)
        self.assertGreater(ssim, 0.9)
        self.assertLess(ssim, 1)
        self.assertFalse(os.path.exists(mergefile))
        self.assertTrue(os.path.exists(expected_diff_file), f"{expected_diff_file} does not exist")

    def test__align_alpha_channel_of_actual_image__no_change(self):
        img_expected = io.imread(self.expected_image)
        img_actual = io.imread(self.actual_image)
        result = self.test_instance._align_alpha_channel_of_actual_image(img_expected, img_actual)
        self.assertTrue(img_actual is result)

    def test__align_alpha_channel_of_actual_image__add_alpha(self):
        img_expected = io.imread(self.expected_image)
        img_actual = io.imread(self.actual_image_rgb)
        result = self.test_instance._align_alpha_channel_of_actual_image(img_expected, img_actual)
        before_color_depth = img_actual.shape[2]
        result_color_depth = result.shape[2]
        self.assertEqual(3, before_color_depth)
        self.assertEqual(4, result_color_depth)

    def test__align_alpha_channel_of_actual_image__remove_alpha(self):
        img_expected = io.imread(self.expected_image_rgb)
        img_actual = io.imread(self.actual_image)
        result = self.test_instance._align_alpha_channel_of_actual_image(img_expected, img_actual)
        before_color_depth = img_actual.shape[2]
        result_color_depth = result.shape[2]
        self.assertEqual(4, before_color_depth)
        self.assertEqual(3, result_color_depth)

    def test_adapt_and_compare_images_merged(self):
        mergefile = os.path.join(self.diffs_dir, "actual_rgba_merged.png")
        self._remove_image_if_it_exists(mergefile)
        ssim = self.test_instance.adapt_and_compare_images(
            expected_screenshot_full_path=self.expected_image,
            actual_screenshot_full_path=self.actual_image,
            output_path=self.diffs_dir,
            diff_formats="color:green",
            append_images=True)
        self.assertGreater(ssim, 0.9)
        self.assertLess(ssim, 1)
        self.assertTrue(os.path.exists(mergefile))
        diff_file = os.path.join(self.diffs_dir, "actual_rgba_green.png")
        self.assertFalse(os.path.exists(diff_file))

    def test_crop_image_file(self):
        file = self._prepare_crop_image()
        self.test_instance.crop_image_file(file,
            location={"x": 10, "y": 10},
            size={"width": 20, "height": 20},
            pixel_ratio=2, viewport_offset=10)
        img = io.imread(file)
        self.assertEqual((40, 40, 4), img.shape)

    def test__pad_images__pad_actual(self):
        img_actual = self._create_img(3, 3)
        img_expected = self._create_img(4, 4)
        img_actual_padded, img_expected_padded = self.test_instance._pad_images(img_actual, img_expected)
        self.assertTrue(img_expected is img_expected_padded)
        self.assertTupleEqual((4, 4, 4), img_actual_padded.shape)
        padded_color = img_actual_padded[3][3]
        expected_color = [255, 0, 0, 255]
        self.assertListEqual(expected_color, padded_color.tolist())

    def test__pad_images__pad_expected(self):
        img_actual = self._create_img(4, 4)
        img_expected = self._create_img(3, 3)
        img_actual_padded, img_expected_padded = self.test_instance._pad_images(img_actual, img_expected)
        self.assertTrue(img_actual is img_actual_padded)
        self.assertTupleEqual((4, 4, 4), img_expected_padded.shape)
        padded_color = img_expected_padded[3][3]
        expected_color = [255, 0, 0, 255]
        self.assertListEqual(expected_color, padded_color.tolist())

    def _create_img(self, width: int, height: int) -> np.ndarray:
        color = [215, 215, 215, 255]
        width = [color for _ in range(width)]
        img = np.array([width for _ in range(height)], dtype=uint8)
        return img_as_ubyte(img)

    def _prepare_crop_image(self):
        # copy a test resource there for further operations
        target = os.path.join(self.crop_dir, "crop.png")
        shutil.copy(self.actual_image, target)
        return target

    def _remove_image_if_it_exists(self, file_path):
        if os.path.exists(file_path):
            os.remove(file_path)

if __name__ == '__main__':
    unittest.main()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import time

from appium.webdriver.webelement import WebElement as AppiumElement
from dataclasses import dataclass
from typing import Any, Callable, List, TypeVar
from getgauge.python import data_store
from selenium.webdriver import Remote
from selenium.webdriver.common.by import By
from selenium.common.exceptions import TimeoutException, WebDriverException
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.remote.webelement import WebElement

from .app_context import AppContext, app_context_key, timeout_key
from .bymapper import ByMapper
from .config import common_config as config
from .report import Report
from .substitute import substitute


T = TypeVar('T')


def find_element(by_param: str, by_value_param: str, immediately=False) -> WebElement | None:
    by = substitute(by_param)
    by_value = substitute(by_value_param)
    marker = get_marker(by, by_value)
    if immediately:
        try:
            return _driver().find_element(*marker)
        except WebDriverException:
            return None
    else:
        return wait_until(lambda driver: driver.find_element(*marker))


def find_elements(by_param: str, by_value_param: str, immediately=False) -> List[WebElement] | None:
    by = substitute(by_param)
    by_value = substitute(by_value_param)
    marker = get_marker(by, by_value)
    try:
        if immediately:
            return _driver().find_elements(*marker)
        else:
            return wait_until(lambda driver: driver.find_elements(*marker))
    except WebDriverException:
        return []


def get_text_from_element(by_param: str, by_value_param: str, immediately=False) -> str | None:
    element = find_element(by_param, by_value_param, immediately)
    if element is None:
        return None
    elif "input" == element.tag_name:
        res = element.get_attribute("value")
        if res is None:
            # workaround for some mobile devices
            res = _driver().execute_script("return arguments[0].value", element)
        if res is not None:
            return res
    return element.text


def find_attribute(by_param: str, by_value_param: str, attribute: str, immediately=False) -> str | bool | None:
    marker = get_marker(substitute(by_param), substitute(by_value_param))
    def _element_attribute(driver: Remote) -> Any:
        """
        This will return the string value of the attribute.
        Empty attributes will return 'true', never an empty string.
        If the attribute does not exist, it will return `False`.
        """
        element = driver.find_element(*marker)
        value = element.get_dom_attribute(attribute)
        return value if value is not None else False
    if immediately:
        try:
            res = _element_attribute(_driver())
            return res if res else None
        except WebDriverException:
            return None
    else:
        return wait_until(_element_attribute)


@dataclass
class ViewportOffset:
    """The position of the element's top left corner in the viewport."""
    top: int
    left: int


def wait_for_idle_element(by: str, by_value: str):
    element = find_element(by, by_value)
    previous_offset = _offset(element)
    def element_stable(_: Remote):
        time.sleep(0.2)
        potentially_moving_element = find_element(by, by_value)
        current_offset = _offset(potentially_moving_element)
        print(f"Got offset: {current_offset}")
        nonlocal previous_offset
        if previous_offset == current_offset:
            return True
        previous_offset = current_offset
        return False
    try:
        wait_until(element_stable)
    except TimeoutException:
        # Best effort approach
        _report().log_debug(f"element {element} keeps moving after scrolling into view")


def wait_until(condition: Callable[[Remote], T], message: str = "") -> T:
    timeout = data_store.scenario.get(timeout_key, config.get_implicit_timeout())
    return WebDriverWait(_driver(), timeout=timeout, poll_frequency=0.25, ignored_exceptions=[WebDriverException])\
        .until(condition, message)


def get_marker(by_string: str, by_value: str) -> tuple[str, str]:
    mapped_by = ByMapper.map_string(by_string)
    if mapped_by == By.ID:
        mapped_by = By.CSS_SELECTOR
        by_value = f"#{by_value}"
    return mapped_by, by_value


def _offset(element: WebElement) -> ViewportOffset:
    if isinstance(element, AppiumElement) and config.is_app_test():
        elem: AppiumElement = element
        location_in_view = elem.location_in_view
        return ViewportOffset(round(location_in_view['y']), round(location_in_view['x']))
    else:
        viewport_offset = _driver().execute_script("return arguments[0].getBoundingClientRect();", element)
        return ViewportOffset(round(viewport_offset['top']), round(viewport_offset['left']))


def _driver() -> Remote:
    app_ctx: AppContext = data_store.spec[app_context_key]
    return app_ctx.driver


def _report() -> Report:
    app_ctx: AppContext = data_store.spec[app_context_key]
    return app_ctx.report



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os

from typing import Optional
from warnings import warn

from ..driver import Browser, OperatingSystem, Platform


def is_app_test() -> bool:
    return os.environ.get("driver_app_location") is not None\
        or os.environ.get("driver_app_package") is not None


def get_app_location() -> Optional[str]:
    return os.environ.get("driver_app_location")


def get_app_package() -> Optional[str]:
    return os.environ.get("driver_app_package")


def get_app_activity() -> Optional[str]:
    return os.environ.get("driver_app_activity")


def is_auto_grant_permissions() -> bool:
    return os.environ.get("driver_app_auto_grant_permissions", "False").lower() in ("true", "1")


def is_debug_log() -> bool:
    return os.environ.get("debug_log", "False").lower() in ("true", "1")


def get_diff_formats() -> str:
    return os.environ.get("diff_formats", "full")


def is_whole_page_screenshot() -> bool:
    return os.environ.get("screenshot_whole_page_no_scroll", "False").lower() in ("true", "1")


def get_time_pattern() -> str:
    return os.environ.get("time_pattern", "%Y-%m-%d_%H-%M-%S")


def get_file_name_pattern() -> str:
    return os.environ.get("filename_pattern", "%{browser}_%{name}.%{ext}")


def get_screenshot_dir() -> str:
    return os.environ.get("screenshot_dir", "screenshots")


def get_actual_screenshot_dir() -> str:
    return os.environ.get("actual_screenshot_dir", "actual_screenshots")


def get_expected_screenshot_dir() -> str:
    return os.environ.get("expected_screenshot_dir", "expected_screenshots")


def get_failure_screenshot_dir() -> str:
    return os.environ.get("failure_screenshot_dir", os.path.join("reports", "html-report", "images"))


def get_browser(default=Browser.FIREFOX) -> Browser:
    config_browser = os.environ.get("driver_browser", default.value)
    return Browser(config_browser)


def get_implicit_timeout(default=5) -> float:
    return float(os.environ.get("driver_implicit_timeout", default))


def get_scroll_wait_time(default=0.7) -> float:
    return float(os.environ.get("driver_scroll_wait_time", default))


def get_operating_system(default=OperatingSystem.MACOS) -> OperatingSystem:
    config_os = os.environ.get("driver_operating_system", default.value)
    return OperatingSystem.parse(config_os)


def get_operating_system_version() -> Optional[str]:
    return os.environ.get("driver_operating_system_version")


def get_page_load_timeout(default=30) -> int:
    return int(os.environ.get("driver_page_load_timeout", default))


def get_platform(default=Platform.LOCAL) -> Platform:
    config_platform = os.environ.get("driver_platform", default.value).lower()
    return Platform(config_platform)


def is_headless() -> bool:
    headless = False
    if "driver_headless" in os.environ:
        warn("property 'driver_headless' is deprecated. Please use 'driver_platform_local_headless' instead")
        headless = os.environ.get("driver_headless").lower() in ("true", "1")
    if "driver_platform_local_headless" in os.environ:
        return os.environ.get("driver_platform_local_headless", "false").lower() in ("true", "1")
    return headless


def get_custom_args() -> list:
    args_prop = os.environ.get("driver_custom_args", "")
    return [arg.strip() for arg in args_prop.split(",") if arg.strip()]


def get_driver_executable_path() -> str | None:
    exec_path = os.environ.get("driver_executable_path")
    return exec_path.strip() if exec_path else None



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import re

from typing import Iterable
from selenium.webdriver.common.keys import Keys


class KeyMapper:

    # {'NULL': '\ue000', 'CANCEL': '\ue001' etc ... }
    _MAPPING = dict((key, getattr(Keys, key)) for key in vars(Keys) if not key.startswith("_"))
    _KEYS = list(_MAPPING.keys())

    @classmethod
    def map_string(cls, key_string: str) -> str:
        return cls._MAPPING[key_string]

    def map_keys(keys_param: str) -> Iterable[str]:
        keys = re.split(r'[,\s]+', keys_param)
        unknown_keys = [k for k in keys if k not in KeyMapper._KEYS]
        assert len(unknown_keys) == 0,\
            "Keys %s unknown.\nUse those instead: %s" % (unknown_keys, KeyMapper._KEYS)
        send_keys = [KeyMapper.map_string(k) for k in keys]
        return send_keys



from os import path
from setuptools import setup


here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    long_description = long_description.replace('./docs/', 'https://github.com/IBM/gauge-web-app-steps/tree/master/docs/')

setup(
    name='gauge-web-app-steps',
    version='0.67',
    description='Provides basic steps for a Gauge project, that runs tests with Selenium and Appium',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/IBM/gauge-web-app-steps',
    author='Tobias Lehmann',
    author_email='derdualist1@gmail.com',
    license='MIT',
    packages=['gauge_web_app_steps', 'gauge_web_app_steps.config', 'gauge_web_app_steps.driver'],
    install_requires=[
        'Appium-Python-Client==5.2.7',
        'getgauge>=0.5.0',
        'numexpr==2.14.1',
        'scikit-image==0.26.0',
        'selenium==4.41.0',
        'webcolors==25.10.0',
        'requests==2.32.5',
    ],
    zip_safe=False
)



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import unittest
from unittest.mock import MagicMock
from gauge_web_app_steps.selector import SelectKey, Selector


class TestSelector(unittest.TestCase):

    def test_to_enum(self):
        key = SelectKey.to_enum("visible text")
        self.assertEqual(SelectKey.VISIBLE_TEXT, key)

    def test_to_enum_invalid(self):
        self.assertRaises(AssertionError, lambda: SelectKey.to_enum())

    def test_select_index(self):
        elm = MagicMock()
        Selector._select(elm, 0)
        self.assertTrue(elm.select_by_index(0))

    def test_select_value(self):
        elm = MagicMock()
        Selector._select(elm, 0, SelectKey.VALUE)
        self.assertTrue(elm.select_by_value(0))

    def test_select_visible_text(self):
        elm = MagicMock()
        Selector._select(elm, 0, SelectKey.VISIBLE_TEXT)
        self.assertTrue(elm.select_by_visible_text(0))


if __name__ == '__main__':
    unittest.main()



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from enum import Enum

from .operating_system import OperatingSystem


class Browser(Enum):

    CHROME = "chrome"
    EDGE = "edge"
    FIREFOX = "firefox"
    INTERNET_EXPLORER = "internet explorer"
    OPERA = "opera"
    SAFARI = "safari"

    def is_supported(self, operating_system: OperatingSystem) -> bool:
        """Checks if given browser is supported by the specified operating system."""
        supported = {
            OperatingSystem.MACOS: [Browser.CHROME, Browser.EDGE, Browser.FIREFOX , Browser.OPERA, Browser.SAFARI],
            OperatingSystem.WINDOWS: [Browser.CHROME, Browser.EDGE, Browser.FIREFOX, Browser.INTERNET_EXPLORER, Browser.OPERA],
            OperatingSystem.LINUX: [Browser.CHROME, Browser.FIREFOX, Browser.OPERA],
            OperatingSystem.IOS: [Browser.FIREFOX, Browser.OPERA, Browser.SAFARI],
            OperatingSystem.ANDROID: [Browser.CHROME, Browser.FIREFOX, Browser.OPERA]
        }
        return self in supported[operating_system]

    def __str__(self) -> str:
        return self.value



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os
import re
import numpy as np
import webcolors

from webcolors import HTML4
from warnings import warn
from skimage import img_as_ubyte
from skimage import color as skimg_color
from skimage import exposure as skimg_exposure
from skimage import io as skimg_io
from skimage import transform as skimg_transform
from skimage import util as skimg_util
from skimage.metrics import structural_similarity as compare_ssim

from .report import Report


class Images(object):
    """
    Functionality around image processing and comparison.
    """

    def __init__(self, report_: Report):
        self.report = report_

    def crop_image_file(
            self,
            screenshot_file_path: str,
            location: dict,
            size: dict,
            pixel_ratio: int,
            viewport_offset: int
    ):
        """
        Crop the specified image by the given dimensions.
        Parameters
        ----------
        screenshot_file_path : str
        location: Dict: {"x": int, "y": int}
            the upper left corner of the rectangle to be cropped out of the image
        size: Dict: {"width": int, "height": int}
            the size of the cropping
        pixel_ratio:
            for Retina displays this is usually 2, else 1
        viewport_offset:
            the offset of the browser's viewport
        """
        img = skimg_io.imread(screenshot_file_path)
        img = self._crop_image(img, location, size, pixel_ratio, viewport_offset)
        skimg_io.imsave(screenshot_file_path, img, check_contrast=False)
        self.report.log_image_info("screenshot {}".format(screenshot_file_path), img)

    def _crop_image(
            self,
            image,
            location,
            size,
            pixel_ratio: int,
            viewport_offset
    ):
        if pixel_ratio != 1:
            self.report.log("pixel ratio is %s" % (pixel_ratio,))
        start_x = int(location["x"]) * pixel_ratio
        start_y = int(location["y"]) * pixel_ratio - viewport_offset
        end_x = start_x + int(size["width"]) * pixel_ratio
        end_y = start_y + int(size["height"]) * pixel_ratio
        self.report.log("cropping image to element dimensions {}x{}".format((end_x - start_x), (end_y - start_y)))
        cropped_image = image[start_y: end_y, start_x: end_x]
        return cropped_image

    def adapt_and_compare_images(
            self,
            expected_screenshot_full_path: str,
            actual_screenshot_full_path: str,
            diff_formats="full",
            append_images=False,
            output_path=""
    ) -> float:
        """
        Calculates the SSIM between 2 images. Does rescaling and padding of the actual image, if necessary.
        Also prints a diff image to the report in different possible formats.
        Parameters
        ----------
        expected_screenshot_full_path : str
        actual_screenshot_full_path : str
        diff_formats : str
            One or more of the following in the same string: gradient, full, (color:red|color:green|color:fuchsia|...).
            The name of the color should a valid HTML color name: https://www.w3.org/TR/html401/types.html#h-6.5.
        append_images: if true, the diff image will be append to the expected screenshot
        output_path : optional path for the diff image, if not set it is equal to the path of the actual screenshot
        """
        img_expected = skimg_io.imread(expected_screenshot_full_path)
        img_actual_raw = skimg_io.imread(actual_screenshot_full_path)
        img_actual = self._align_alpha_channel_of_actual_image(img_expected, img_actual_raw)
        channel_axis = self._channel_axis(img_actual)
        self.report.log_debug(f"actual channel_axis: {channel_axis}")
        self.report.log_debug(f"expected channel_axis: {self._channel_axis(img_expected)}")
        img_actual = self._rescale_image(img_actual, img_expected, channel_axis)
        img_actual, img_expected = self._pad_images(img_actual, img_expected)
        if img_actual is not img_actual_raw:
            self.report.log_debug("Overwriting actual image after rescaling and padding")
            skimg_io.imsave(actual_screenshot_full_path, img_actual, check_contrast=False)
        self.report.log_image_info("actual", img_actual)
        self.report.log_image_info("expected", img_expected)
        img_list = []
        if append_images:
            img_list.append(img_expected)
        ssim, diff_images = self._compute_ssim_and_diff(img_actual, img_expected, diff_formats, channel_axis=channel_axis)
        if ssim < 1.0:
            self._save_diff_image(expected_screenshot_full_path, actual_screenshot_full_path, output_path,
                                  diff_images, img_list)
        self.report.log_debug("SSIM: {}\n".format(ssim))
        return ssim

    def _align_alpha_channel_of_actual_image(self, img_expected: np.ndarray, img_actual: np.ndarray) -> np.ndarray:
        expected_has_alpha = self._img_has_alpha(img_expected)
        actual_has_alpha = self._img_has_alpha(img_actual)
        if expected_has_alpha and not actual_has_alpha:
            self.report.log_debug("Adding alpha channel to actual image")
            rgba = np.insert(
                    img_actual,
                    3, # index in the color array [r,g,b,a]. index 3 -> a.
                    255, # value to insert at the index above: 255 is fully opaque
                    axis=2, # 0: height, 1: width, 2: color
            )
            return rgba
        elif not expected_has_alpha and actual_has_alpha:
            self.report.log_debug("Removing alpha channel from actual image")
            img_rgb = skimg_color.rgba2rgb(img_actual)
            return img_as_ubyte(img_rgb)
        else:
            return img_actual

    def _img_has_alpha(self, img: np.ndarray):
        height, width, color = img.shape
        # 4 bytes of color information indicates RGBa, 3 would be without.
        return color == 4

    def _channel_axis(self, img: np.ndarray):
        channel_axis = img.ndim - 1
        return channel_axis  # 2 dimensions: b/w, 3: color

    def _rescale_image(
            self,
            img,
            img_reference,
            channel_axis
    ):
        rescale_ratio = self._compute_rescale_ratio(img, img_reference)
        if rescale_ratio == 1.0:
            return img
        self.report.log("rescaling image by ratio {} to fit expected image size".format(rescale_ratio))
        img_rescaled = skimg_transform.rescale(
            img,
            rescale_ratio,
            channel_axis=channel_axis,
            anti_aliasing=True
        )
        # skimage uses different internal representations for an image.
        # https://scikit-image.org/docs/dev/user_guide/data_types.html
        # The rescale function returns an image with a different data type, so we convert it back.
        return img_as_ubyte(img_rescaled)

    def _compute_rescale_ratio(
            self,
            img,
            img_reference
    ):
        actual_height = len(img)
        reference_height = len(img_reference)
        if actual_height == reference_height:
            return 1.0
        actual_width = len(img[0])
        reference_width = len(img_reference[0])
        if actual_width == reference_width:
            return 1.0
        height_ratio = 1.0 * reference_height / actual_height
        width_ratio = 1.0 * reference_width / actual_width
        return height_ratio if height_ratio < width_ratio else width_ratio

    def _pad_images(
            self,
            img,
            img_ref
    ) -> tuple[np.ndarray, np.ndarray]:
        pad_bottom = len(img_ref) - len(img)
        pad_right = len(img_ref[0]) - len(img[0])
        if pad_bottom == pad_right == 0:
            return img, img_ref
        padded_ref_image = img_ref
        padded_image = img
        if pad_bottom < 0 or pad_right < 0:
            ref_pad_bottom = -min(pad_bottom, 0)
            ref_pad_right = -min(pad_right, 0)
            self.report.log(f"reference screenshot needs padding to match actual image size. padding bottom: {ref_pad_bottom}, padding right: {ref_pad_right}")
            padded_ref_image = self._pad_image(img_ref, ref_pad_bottom, ref_pad_right)
        if pad_bottom > 0 or pad_right > 0:
            act_pad_bottom = max(pad_bottom, 0)
            act_pad_right = max(pad_right, 0)
            self.report.log(f"screenshot needs padding to match expected image size. padding bottom: {act_pad_bottom}, padding right: {act_pad_right}")
            padded_image = self._pad_image(img, act_pad_bottom, act_pad_right)
        return padded_image, padded_ref_image

    def _pad_image(self, img: np.ndarray, pad_bottom: int, pad_right: int) -> np.ndarray:
        height, width, color = img.shape
        red = [255, 0, 0, 255] if self._img_has_alpha(img) else [255, 0, 0]
        red = np.array(red, dtype=np.uint8)
        right_pad_img = np.zeros((height, pad_right, color), dtype=np.uint8) + red
        padded = np.concatenate((img, right_pad_img), axis=1, dtype=np.uint8)
        bottom_pad_img = np.zeros((pad_bottom, width + pad_right, color), dtype=np.uint8) + red
        padded = np.concatenate((padded, bottom_pad_img), axis=0, dtype=np.uint8)
        return padded

    def _compute_ssim_and_diff(
            self,
            img_expected,
            img_actual,
            diff_formats,
            channel_axis
    ):
        ssim = None
        self.report.log_debug(f"using {diff_formats} to compare images")
        diff_images = {}
        if "gradient" in diff_formats and "full" in diff_formats:
            ssim, gradient, full = compare_ssim(img_actual, img_expected, channel_axis=channel_axis, gradient=True, full=True, data_range=255)
            diff_images["gradient"] = self._ssim_img_to_ubyte(gradient)
            diff_images["full"] = self._ssim_img_to_ubyte(full)
        elif "gradient" in diff_formats:
            ssim, gradient = compare_ssim(img_actual, img_expected, channel_axis=channel_axis, gradient=True, data_range=255)
            diff_images["gradient"] = self._ssim_img_to_ubyte(gradient)
        elif "full" in diff_formats:
            ssim, full = compare_ssim(img_actual, img_expected, channel_axis=channel_axis, full=True, data_range=255)
            diff_images["full"] = self._ssim_img_to_ubyte(full)
        if ssim is None:
            ssim = compare_ssim(img_actual, img_expected, channel_axis=channel_axis, data_range=255)
        if ssim < 1.0 and "red" in diff_formats:
            warn("The diff_format 'red' is deprecated. please use a key value pair: 'color=red'", DeprecationWarning)
            red = self._diff_images_color(img_expected, img_actual, "red")
            diff_images["red"] = red
        if ssim < 1.0 and "color:" in diff_formats:
            color_name = re.search("color:([a-z]*)", diff_formats).group(1)
            colored = self._diff_images_color(img_expected, img_actual, color_name)
            diff_images[color_name] = colored
        return ssim, diff_images

    def _ssim_img_to_ubyte(self, img: np.ndarray) -> np.ndarray:
        """ the returned image of *compare_ssim* has a weird format and must be converted back to uint8. """
        img = img * 0.5
        img = img.clip(min=-1.0, max=1.0)
        return img_as_ubyte(img)

    def _diff_images_color(
            self,
            img_expected,
            img_actual,
            color_name
    ):
        """
        Creates a diff image between two images.
        The diff will show any color differences by highlighting with the given color and reducing other colors.
        The name of the color should be HTML compliant.
        """
        color = list(webcolors.name_to_rgb(color_name, HTML4))
        self.report.log_debug("Expected: tranform to RGB float")
        img_expected_rgb = self._transform_to_rgb_float(img_expected)
        self.report.log_debug("Actual: tranform to RGB float")
        img_actual_rgb = self._transform_to_rgb_float(img_actual)
        img_diff = skimg_color.deltaE_cie76(img_expected_rgb, img_actual_rgb)  # Euclidean delta
        img_diff = skimg_color.gray2rgb(img_diff)
        img_diff = img_diff * color
        img_diff = img_expected_rgb + img_diff
        # Normalize the max pixel value, simultaneously graying out the parts of the original image.
        # then transform to 8bit colors
        img_diff = skimg_exposure.rescale_intensity(img_diff, in_range=(0., 2.), out_range=(0, 255))
        return img_diff.astype(np.uint8)

    def _transform_to_rgb_float(self, img):
        """
        Cut the alpha channel from an image, if present, and transform the data representation to float.
        Each channel will range between 0.0 and 1.0.
        """
        black = [0, 0, 0]
        if self._img_has_alpha(img):
            self.report.log_debug("Image has Alpha")
            return skimg_color.rgba2rgb(img, background=black)
        else:
            self.report.log_debug("Image has no Alpha")
            return skimg_util.img_as_float(img)

    def _save_diff_image(
            self,
            expected_screenshot_full_path,
            actual_screenshot_full_path,
            output_path,
            diff_images,
            img_list
    ):
        path = self._determine_target_path(actual_screenshot_full_path, output_path)
        for diff_format, diff_img in diff_images.items():
            if len(img_list) > 0:
                # append the images to the list to create a big one for visual comparing
                img_list.append(diff_img)
            else:
                # save diff images immediately
                diff_path = self._create_target_filename(path, diff_format)
                skimg_io.imsave(diff_path, diff_img, check_contrast=False)
                self.report.log_image(diff_path, f"Created {diff_format} diff for {expected_screenshot_full_path}")
        self._create_horizontal_aligned_diff(expected_screenshot_full_path, path, img_list)

    def _create_horizontal_aligned_diff(
            self,
            expected_screenshot_full_path,
            path,
            img_list
    ):
        """
        Creates a horizontally stacked image from expected image and diffs.
        """
        if len(img_list) > 1:
            diff_path = self._create_target_filename(path, "merged")
            reshaped = [self._transform_to_rgb_float(i) for i in img_list]
            merged = np.concatenate(reshaped, axis=1)
            merged = img_as_ubyte(merged)
            skimg_io.imsave(diff_path, merged, check_contrast=False)
            self.report.log_image(diff_path, f"Created merged diff for {expected_screenshot_full_path}")

    def _determine_target_path(
            self,
            actual_screenshot_full_path,
            output_path
    ):
        if not output_path or output_path.isspace():
            # output_path is None, empty or contains only whitespaces
            return actual_screenshot_full_path
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        filename = os.path.basename(actual_screenshot_full_path)
        return os.path.join(output_path, filename)

    def _create_target_filename(
            self,
            path,
            suffix
    ):
        return "{}_{}.png".format(path[:-len(".png")], suffix)



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

from __future__ import annotations
import os
from abc import ABC, abstractmethod
from appium.options.android import UiAutomator2Options
from appium.options.common import AppiumOptions
from appium.options.ios import XCUITestOptions
from appium.webdriver.webdriver import WebDriver as MobileRemote
from appium.webdriver import Remote as NativeRemote
from selenium.webdriver.remote.webdriver import WebDriver as Remote
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.chrome.webdriver import WebDriver as Chrome
from selenium.webdriver.common.options import ArgOptions
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.edge.service import Service as EdgeService
from selenium.webdriver.edge.webdriver import WebDriver as Edge
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.firefox.webdriver import WebDriver as Firefox
from selenium.webdriver.ie.options import Options as IeOptions
from selenium.webdriver.ie.service import Service as IeService
from selenium.webdriver.ie.webdriver import WebDriver as Ie
from selenium.webdriver.safari.options import Options as SafariOptions
from selenium.webdriver.safari.service import Service as SafariService
from selenium.webdriver.safari.webdriver import WebDriver as Safari

from ..config import common_config as config
from ..config import local_config
from ..config import saucelabs_config
from .browsers import Browser
from .operating_system import OperatingSystem, SaucelabsOperatingSystem


class DriverFactory(ABC):

    """
    Abstract parent factory for all driver factories.
    """

    @staticmethod
    def create_driver_factory(spec_name: str, suite_id: str) -> DriverFactory:
        if config.get_platform().is_remote():
            return SaucelabsDriverFactory(spec_name, suite_id)
        elif config.get_platform().is_local():
            return LocalDriverFactory()
        else:
            raise RuntimeError("Given driver options are not supported")

    @abstractmethod
    def create_driver(self) -> Remote:
        pass

    def _create_browser_options(self) -> ArgOptions | AppiumOptions:
        if config.get_operating_system().is_mobile():
            operating_system = config.get_operating_system()
            return {
                OperatingSystem.ANDROID: UiAutomator2Options,
                OperatingSystem.IOS: XCUITestOptions,
            }[operating_system]()
        else:
            browser = config.get_browser()
            return {
                Browser.CHROME: ChromeOptions,
                Browser.EDGE: EdgeOptions,
                Browser.FIREFOX: FirefoxOptions,
                Browser.INTERNET_EXPLORER: IeOptions,
                Browser.OPERA: ChromeOptions,
                Browser.SAFARI: SafariOptions,
            }[browser]()


class LocalDriverFactory(DriverFactory):

    """
    This factory creates drivers for a browser which is installed on the machine of the user.
    """

    def create_driver(self) -> Remote:
        """Creates and returns a driver for local environment."""
        if config.get_operating_system().is_desktop():
            return self._create_desktop_driver()
        elif config.get_operating_system().is_mobile():
            return self._create_mobile_driver()
        else:
            raise RuntimeError(f"Operating system {config.get_operating_system()} is not supported")

    def _create_desktop_driver(self) -> Remote:
        browser = config.get_browser()
        operating_system = config.get_operating_system()
        assert browser.is_supported(operating_system), f"Browser {browser} not supported by {operating_system}."
        browser_options = self._create_options()
        executable_path = config.get_driver_executable_path()
        service = {
            Browser.CHROME: lambda: ChromeService(executable_path=executable_path),
            Browser.EDGE: lambda: EdgeService(executable_path=executable_path),
            Browser.FIREFOX: lambda: FirefoxService(executable_path=executable_path, log_output=self._firefox_log_path()),
            Browser.INTERNET_EXPLORER: lambda: IeService(executable_path=executable_path),
            Browser.OPERA: lambda: ChromeService(executable_path=executable_path),
            Browser.SAFARI: lambda: SafariService() # Driver executable for Mac/Safari comes pre-installed
        }[browser]()
        driver: Remote = None
        driver = {
            Browser.CHROME: lambda: Chrome(service=service, options=browser_options),
            Browser.EDGE: lambda: Edge(service=service, options=browser_options),
            Browser.FIREFOX: lambda: Firefox(service=service, options=browser_options),
            Browser.INTERNET_EXPLORER: lambda: Ie(service=service, options=browser_options),
            Browser.OPERA: lambda: Chrome(service=service, options=browser_options),
            Browser.SAFARI: lambda: Safari(service=service, options=browser_options)
        }[browser]()
        driver.implicitly_wait(config.get_implicit_timeout())
        driver.set_page_load_timeout(config.get_page_load_timeout())
        return driver

    def _create_mobile_driver(self) -> MobileRemote:
        operating_system = config.get_operating_system()
        browser = config.get_browser()
        assert browser.is_supported(operating_system), f"Browser {browser} not supported by {operating_system}."
        capabilities = {
            OperatingSystem.IOS: self._create_ios_capabilities,
            OperatingSystem.ANDROID: self._create_android_capabilities,
        }[operating_system]()
        options = self._create_browser_options()
        capabilities_options = options.load_capabilities(capabilities)
        driver = MobileRemote(local_config.get_mobile_appium_server_url(), options=capabilities_options)
        driver.implicitly_wait(config.get_implicit_timeout())
        driver.set_page_load_timeout(config.get_page_load_timeout())
        return driver

    def _create_options(self) -> ArgOptions:
        browser = config.get_browser()
        options: ArgOptions = None
        if browser == Browser.CHROME:
            options = self._create_chrome_options()
        elif browser == Browser.FIREFOX:
            options = self._create_firefox_options()
        elif browser == Browser.OPERA:
            options = self._create_chrome_options()
            # following is needed for Opera to run
            options.add_experimental_option('w3c', True)
        else:
            options = self._create_browser_options()
        for arg in config.get_custom_args():
            options.add_argument(arg)
        return options

    def _create_firefox_options(self) -> FirefoxOptions:
        options = FirefoxOptions()
        if config.is_headless():
            options.add_argument('-headless')
        return options

    def _create_chrome_options(self) -> ChromeOptions:
        options = ChromeOptions()
        options.add_argument('--no-sandbox')
        options.add_argument('--hide-scrollbars')
        options.add_argument('--disable-extensions')
        if config.is_headless():
            # to run chrome headless in a docker container without any GUI environment
            options.add_argument('--headless')
            options.add_argument('--disable-dev-shm-usage')
            options.add_argument('--start-maximized')
        return options

    def _firefox_log_path(self) -> str:
        project_root = os.environ.get("GAUGE_PROJECT_ROOT")
        return os.path.join(project_root, "logs", "geckodriver.log")

    def _create_android_capabilities(self) -> dict:
        capabilities = {
            "deviceName": local_config.get_mobile_device_name(),
            "platformName": OperatingSystem.ANDROID.value,
            "platformVersion": config.get_operating_system_version(),
            "browserName": config.get_browser().value,
            "nativeWebScreenshot": "true",
            "goog:chromeOptions": {
                "w3c": True,
                "extensions": []
            }
        }
        if local_config.get_mobile_real_device():
            # a real device needs a UDID instead of a name
            capabilities["udid"] = local_config.get_mobile_device_udid()
        custom_args = config.get_custom_args()
        if custom_args:
            capabilities['goog:chromeOptions']['args'] = custom_args
        return capabilities

    def _create_ios_capabilities(self) -> dict:
        capabilities = {
            "automationName": "XCUITest",
            "deviceName": local_config.get_mobile_device_name(),
            "platformName": OperatingSystem.IOS.value,
            "platformVersion": config.get_operating_system_version(),
            "browserName": Browser.SAFARI.value
        }
        custom_args = config.get_custom_args()
        if custom_args:
            # not sure, if this works with safari. Documentation is scarce
            capabilities['safariOptions'] = {}
            capabilities['safariOptions']['args'] = custom_args
        return capabilities


class RemoteDriverFactory(DriverFactory):

    """
    This factory creates remote drivers for browsers. It is not bound to any specific proved of an environment.
    """

    def create_driver(self) -> Remote:
        options = self._create_options()
        return Remote(command_executor=local_config.get_mobile_appium_server_url(), options=options)

    def _create_options(self) -> ArgOptions:
        browser = config.get_browser()
        headless = config.is_headless()
        options = self._create_browser_options()
        if browser == Browser.CHROME or browser == Browser.OPERA:
            options.add_argument('--no-sandbox')
            options.add_argument('--hide-scrollbars')
            options.add_argument('--disable-extensions')
            if headless:
                # to run chrome headless in a docker container without any GUI environment
                options.add_argument('--headless')
                options.add_argument('--disable-dev-shm-usage')
                options.add_argument('--start-maximized')
            if browser == Browser.OPERA:
                options.add_experimental_option('w3c', True)
        elif browser == Browser.EDGE and headless:
            options.add_argument('--headless')
        elif browser == Browser.FIREFOX and headless:
            options.add_argument('-headless')
        return options


class SaucelabsDriverFactory(DriverFactory):
    """
    This factory creates drivers for the remote Saucelabs environment.
    """

    def __init__(self, spec_name: str, suite_id: str) -> None:
        self.spec_name = spec_name
        self.suite_id = suite_id

    def create_driver(self) -> Remote:
        """Creates and returns a driver for https://saucelabs.com/"""
        operating_system = config.get_operating_system()
        if operating_system.is_desktop():
            return self._create_desktop_driver()
        elif operating_system.is_mobile():
            return self._create_mobile_driver()
        else:
            raise RuntimeError(f"Operating system {operating_system} is not supported")

    def _create_desktop_driver(self) -> Remote:
        operating_system = config.get_operating_system()
        browser = config.get_browser()
        assert browser.is_supported(operating_system), f"Browser {browser} not supported by {operating_system}."
        browser_options = self._create_browser_options()
        browser_options.platform_name = self._get_platform_name(operating_system, config.get_operating_system_version())
        browser_options.browser_version = saucelabs_config.get_browser_version()
        sauce_options = self._get_sauce_options()
        browser_options.set_capability("sauce:options", sauce_options)
        for arg in config.get_custom_args():
            browser_options.add_argument(arg)
        driver = Remote(saucelabs_config.get_executor(), options=browser_options)
        driver.implicitly_wait(config.get_implicit_timeout())
        driver.set_page_load_timeout(config.get_page_load_timeout())
        return driver

    def _create_mobile_driver(self) -> MobileRemote | NativeRemote:
        operating_system = config.get_operating_system()
        capabilities = {
            'platformName': operating_system.value,
            'appium:deviceName': saucelabs_config.get_device_name(),
            'appium:platformVersion': config.get_operating_system_version(),
            'sauce:options': self._get_sauce_options()
        }
        options = self._create_browser_options()
        app_test = config.is_app_test()
        if app_test:
            app_location = config.get_app_location()
            if app_location:
                capabilities['appium:app'] = app_location
            app_package = config.get_app_package()
            if app_package:
                capabilities['appium:appPackage'] = app_package
            app_activity = config.get_app_activity()
            if app_activity:
                capabilities['appium:appActivity'] = app_activity
            capabilities['enableMultiWindows'] = True
        else:
            browser = config.get_browser()
            capabilities['browserName'] = browser.value
        auto_grant_permissions = config.is_auto_grant_permissions() and app_test
        custom_args = config.get_custom_args()
        if operating_system == OperatingSystem.ANDROID:
            capabilities['appium:automationName'] = 'UiAutomator2'
            capabilities["goog:chromeOptions"] = {
                'w3c': True,
                'extensions': []
            }
            if custom_args:
                capabilities['goog:chromeOptions']['args'] = custom_args
            if auto_grant_permissions:
                capabilities['appium:autoGrantPermissions'] = True
        elif operating_system == OperatingSystem.IOS:
            capabilities["appium:automationName"] = "XCUITest"
            capabilities["appium:includeSafariInWebviews"] = True
            capabilities["safariAllowPopups"] = True
            if custom_args:
                # not sure, if this works with safari. Documentation is scarce
                capabilities['safariOptions'] = {}
                capabilities['safariOptions']['args'] = custom_args
            if auto_grant_permissions:
                capabilities['appium:autoAcceptAlerts'] = True
        capabilities_options = options.load_capabilities(capabilities)
        if app_test:
            driver = NativeRemote(saucelabs_config.get_executor(), options=capabilities_options, keep_alive=True)
            if operating_system == OperatingSystem.IOS:
                driver.set_page_load_timeout(config.get_page_load_timeout())
        else:
            driver = MobileRemote(saucelabs_config.get_executor(), options=capabilities_options)
            driver.set_page_load_timeout(config.get_page_load_timeout())
        driver.implicitly_wait(config.get_implicit_timeout())
        return driver

    def _get_platform_name(self, operating_system: OperatingSystem, operating_system_version: str) -> str:
        if operating_system == OperatingSystem.WINDOWS:
            return self._get_windows_platform_name(operating_system_version)
        elif operating_system == OperatingSystem.MACOS:
            return self._get_macos_platform_name(operating_system_version)
        else:
            raise RuntimeError(f"Cannot determine platform name for {operating_system}")

    def _get_windows_platform_name(self, operating_system_version: str) -> str:
        version = {
            "11": SaucelabsOperatingSystem.WINDOWS_11,
            "10": SaucelabsOperatingSystem.WINDOWS_10,
            "8.1": SaucelabsOperatingSystem.WINDOWS_8_1,
            "8": SaucelabsOperatingSystem.WINDOWS_8,
            "7": SaucelabsOperatingSystem.WINDOWS_7,
        }
        return version.get(operating_system_version, SaucelabsOperatingSystem.WINDOWS_DEFAULT).value

    def _get_macos_platform_name(self, operating_system_version: str) -> str:
        if operating_system_version is None:
            return SaucelabsOperatingSystem.MACOS_DEFAULT.value
        operating_system_version = operating_system_version.lower()
        versions = {
            "14": SaucelabsOperatingSystem.MACOS_14,
            "sonoma": SaucelabsOperatingSystem.MACOS_14,
            "13": SaucelabsOperatingSystem.MACOS_13,
            "ventura": SaucelabsOperatingSystem.MACOS_13,
            "12": SaucelabsOperatingSystem.MACOS_12,
            "monterey": SaucelabsOperatingSystem.MACOS_12,
            "11": SaucelabsOperatingSystem.MACOS_11,
            "big sur": SaucelabsOperatingSystem.MACOS_11,
            "big_sur": SaucelabsOperatingSystem.MACOS_11,
            "big-sur": SaucelabsOperatingSystem.MACOS_11,
            "10.15": SaucelabsOperatingSystem.MACOS_10_15,
            "10": SaucelabsOperatingSystem.MACOS_10_15,
            "catalina": SaucelabsOperatingSystem.MACOS_10_15,
            "10.14": SaucelabsOperatingSystem.MACOS_10_14,
            "mojave": SaucelabsOperatingSystem.MACOS_10_14,
            "10.13": SaucelabsOperatingSystem.MACOS_10_13,
            "high sierra": SaucelabsOperatingSystem.MACOS_10_13,
            "high_sierra": SaucelabsOperatingSystem.MACOS_10_13,
            "high-sierra": SaucelabsOperatingSystem.MACOS_10_13,
            "10.12": SaucelabsOperatingSystem.MACOS_10_12,
            "sierra": SaucelabsOperatingSystem.MACOS_10_12,
            "10.11": SaucelabsOperatingSystem.MACOS_10_11,
            "el capitan": SaucelabsOperatingSystem.MACOS_10_11,
            "el_capitan": SaucelabsOperatingSystem.MACOS_10_11,
            "el-capitan": SaucelabsOperatingSystem.MACOS_10_11,
            "10.10": SaucelabsOperatingSystem.MACOS_10_10,
            "yosemite": SaucelabsOperatingSystem.MACOS_10_10,
        }
        version = versions.get(operating_system_version)
        return operating_system_version if version is None else version.value

    def _get_sauce_options(self) -> dict:
        sauce_options = {
            "username": saucelabs_config.get_sauce_user_name(),
            "accessKey": saucelabs_config.get_sauce_access_key(),
            "networkCapture": True,
        }
        if config.is_app_test():
            sauce_options["resigningEnabled"] = True
        if saucelabs_config.is_extended_debugging_enabled():
            sauce_options["extendedDebugging"] = True
        operating_system = config.get_operating_system()
        appium_version = saucelabs_config.get_appium_version()
        if operating_system and operating_system.is_mobile() and appium_version:
            sauce_options["appiumVersion"] = appium_version
        tunnel_name = saucelabs_config.get_tunnel_name()
        if tunnel_name and saucelabs_config.is_sauce_tunnel_active():
            sauce_options["tunnelName"] = tunnel_name
        custom_test_title = saucelabs_config.get_test_title()
        spec_name = self.spec_name
        sauce_options["name"] = f"{spec_name} - {custom_test_title}" if custom_test_title else spec_name
        build = saucelabs_config.get_build()
        if build:
            sauce_options["build"] = build
        if saucelabs_config.is_device_cached():
            sauce_options["cacheId"] = self.suite_id
        return sauce_options



#
# Copyright IBM Corp. 2019-
# SPDX-License-Identifier: MIT
#

import os
import errno

from datetime import datetime
from string import Template

from .config import common_config as config


class ImagePath(object):

    def __init__(self, browser: str, headless: bool):
        self.browser_name = browser
        self.headless = headless

    def create_failure_screenshot_file_path(self) -> str:
        now = datetime.now()
        date_str = now.strftime("%Y-%m-%d_%H-%M-%S-%f")

        failure_screenshot_dir = self._create_screenshot_dir(config.get_failure_screenshot_dir())
        screenshot_file = "fail_{}.png".format(date_str)
        screenshot_file_path = os.path.join(failure_screenshot_dir, screenshot_file)
        return screenshot_file_path

    def create_screenshot_file_path(self, image_file_name: str) -> str:
        """
        Creates a new file name in the screenshot directory.
        The directory will be created if it does not exist yet.
        """
        file_name = image_file_name if image_file_name.endswith(".png") else "{}.png".format(image_file_name)
        return "/".join((self._screenshot_dir(), self._filename(file_name),))

    def create_actual_screenshot_file_path(self, image_file_name: str, running_number=None) -> str:
        """
        Creates a new file name in the actual_screenshot directory.
        The directory will be created if it does not exist yet.
        """
        actual_screenshot_dir = self._actual_screenshot_dir()
        actual_screenshot_full_path = self._create_screenshot_file_path(actual_screenshot_dir, image_file_name, running_number)
        return actual_screenshot_full_path

    def create_expected_screenshot_file_path(self, image_file_name: str, running_number=None) -> str:
        """
        Creates a new file name in the expected_screenshot directory.
        The directory will be created if it does not exist yet.
        """
        expected_screenshot_dir = self._expected_screenshot_dir()
        expected_screenshot_full_path = self._create_screenshot_file_path(expected_screenshot_dir, image_file_name, running_number)
        return expected_screenshot_full_path

    def _create_screenshot_file_path(self, screenshot_dir: str, image_file_name: str, running_number=None) -> str:
        filename_without_ext = image_file_name if not image_file_name.endswith(".png") else image_file_name[:-len(".png")]
        if running_number is None:
            file_name_parts = [self.browser_name, "_", filename_without_ext, ".png"]
        else:
            file_name_parts = [self.browser_name, "_", filename_without_ext, "_", str(running_number), ".png"]
        if self.headless:
            file_name_parts.insert(0, "headless_")
        file_name = "".join(file_name_parts)
        screenshot_full_path = os.path.join(screenshot_dir, file_name)
        return screenshot_full_path

    def _time_str(self) -> str:
        time_pattern = config.get_time_pattern()
        now = datetime.now()
        date_str = now.strftime(time_pattern)
        return date_str

    def _filename(self, name) -> str:
        pattern = config.get_file_name_pattern()
        class FileTemplate(Template):
            delimiter = "%"
        template = FileTemplate(pattern)
        subs = template.safe_substitute({
                "name"    : name if not name.endswith(".png") else name[:-len(".png")],
                "ext"     : "png",
                "browser" : self.browser_name,
                "time"    : self._time_str(),
        })
        return subs if subs.endswith(".png") else "{}.png".format(subs)

    def _screenshot_dir(self) -> str:
        return self._create_screenshot_dir(config.get_screenshot_dir())

    def _actual_screenshot_dir(self) -> str:
        return self._create_screenshot_dir(config.get_actual_screenshot_dir())

    def _expected_screenshot_dir(self) -> str:
        return self._create_screenshot_dir(config.get_expected_screenshot_dir())

    def _create_screenshot_dir(self, screenshot_dir) -> str:
        if not os.path.isabs(screenshot_dir):
            project_root = os.environ.get("GAUGE_PROJECT_ROOT")
            screenshot_dir = os.path.join(project_root, screenshot_dir)
        screenshot_dir = os.path.abspath(screenshot_dir)
        return self._create_dir(screenshot_dir)

    def _create_dir(self, dir_) -> str:
        try:
            os.makedirs(dir_)
        except OSError as e:
            # the directory might already exist. If not it is created.
            if e.errno != errno.EEXIST:
                raise
            pass
        return dir_
